# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
from PySide6 import QtCore, QtWidgets, QtGui

from onionshare_cli.common import Common
from onionshare_cli.web import Web

from .threads import CompressThread
from .. import Mode
from ..file_selection import FileSelection
from ..history import History, ToggleHistory, ShareHistoryItem
from .... import strings
from ....widgets import MinimumSizeWidget
from ....gui_common import GuiCommon


class ShareMode(Mode):
    """
    Parts of the main window UI for sharing files.
    """

    def init(self):
        """
        Custom initialization for ReceiveMode.
        """
        # Threads start out as None
        self.compress_thread = None

        # Create the Web object
        self.web = Web(self.common, True, self.settings, "share")

        # Settings
        self.autostop_sharing_checkbox = QtWidgets.QCheckBox()
        self.autostop_sharing_checkbox.clicked.connect(
            self.autostop_sharing_checkbox_clicked
        )
        self.autostop_sharing_checkbox.setText(
            strings._("mode_settings_share_autostop_sharing_checkbox")
        )
        if self.settings.get("share", "autostop_sharing"):
            self.autostop_sharing_checkbox.setCheckState(QtCore.Qt.Checked)
        else:
            self.autostop_sharing_checkbox.setCheckState(QtCore.Qt.Unchecked)

        self.mode_settings_widget.mode_specific_layout.addWidget(
            self.autostop_sharing_checkbox
        )

        # File selection
        self.file_selection = FileSelection(
            self.common,
            "images/{}_mode_share.png".format(self.common.gui.color_mode),
            strings._("gui_new_tab_share_button"),
            self,
        )
        if self.filenames:
            for filename in self.filenames:
                self.file_selection.file_list.add_file(filename)

        # Set title placeholder
        self.mode_settings_widget.title_lineedit.setPlaceholderText(
            strings._("gui_tab_name_share")
        )

        # Server status
        self.server_status.set_mode("share", self.file_selection)
        self.server_status.server_started.connect(self.file_selection.server_started)
        self.server_status.server_stopped.connect(self.file_selection.server_stopped)
        self.server_status.server_stopped.connect(self.update_primary_action)
        self.server_status.server_canceled.connect(self.file_selection.server_stopped)
        self.server_status.server_canceled.connect(self.update_primary_action)
        self.file_selection.file_list.files_updated.connect(self.server_status.update)
        self.file_selection.file_list.files_updated.connect(self.update_primary_action)
        # Tell server_status about web, then update
        self.server_status.web = self.web
        self.server_status.update()

        # Filesize warning
        self.filesize_warning = QtWidgets.QLabel()
        self.filesize_warning.setWordWrap(True)
        self.filesize_warning.setStyleSheet(
            self.common.gui.css["share_filesize_warning"]
        )
        self.filesize_warning.hide()

        # Download history
        self.history = History(
            self.common,
            QtGui.QPixmap.fromImage(
                QtGui.QImage(
                    GuiCommon.get_resource_path("images/share_icon_transparent.png")
                )
            ),
            strings._("gui_share_mode_no_files"),
            strings._("gui_all_modes_history"),
        )
        self.history.hide()

        # Info label
        self.info_label = QtWidgets.QLabel()
        self.info_label.hide()

        # Delete all files button
        self.remove_all_button = QtWidgets.QPushButton(
            strings._("gui_file_selection_remove_all")
        )
        self.remove_all_button.setFlat(True)
        self.remove_all_button.setStyleSheet(
            self.common.gui.css["share_delete_all_files_button"]
        )
        self.remove_all_button.clicked.connect(self.delete_all)
        self.remove_all_button.hide()

        # Toggle history
        self.toggle_history = ToggleHistory(
            self.common,
            self,
            self.history,
            QtGui.QIcon(GuiCommon.get_resource_path(f"images/{self.common.gui.color_mode}_history_icon_toggle.svg")),
            QtGui.QIcon(
                GuiCommon.get_resource_path(f"images/{self.common.gui.color_mode}_history_icon_toggle_selected.svg")
            ),
        )

        # Top bar
        top_bar_layout = QtWidgets.QHBoxLayout()
        top_bar_layout.addWidget(self.info_label)
        top_bar_layout.addStretch()
        top_bar_layout.addWidget(self.remove_all_button)
        top_bar_layout.addWidget(self.toggle_history)

        # Primary action layout
        self.primary_action_layout.addWidget(self.filesize_warning)
        self.primary_action.hide()
        self.update_primary_action()

        # Status bar, zip progress bar
        self._zip_progress_bar = None

        # Main layout
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.addLayout(top_bar_layout)
        self.main_layout.addLayout(self.file_selection)
        self.main_layout.addWidget(self.primary_action, stretch=1)
        self.main_layout.addWidget(self.server_status)
        self.main_layout.addWidget(MinimumSizeWidget(700, 0))

        # Column layout
        self.column_layout = QtWidgets.QHBoxLayout()
        self.column_layout.addLayout(self.main_layout)
        self.column_layout.addWidget(self.history, stretch=1)

        # Content layout
        self.content_layout.addLayout(self.column_layout)

        # Always start with focus on file selection
        self.file_selection.setFocus()

    def get_type(self):
        """
        Returns the type of mode as a string (e.g. "share", "receive", etc.)
        """
        return "share"

    def autostop_sharing_checkbox_clicked(self):
        """
        Save autostop sharing setting to the tab settings
        """
        self.settings.set(
            "share", "autostop_sharing", self.autostop_sharing_checkbox.isChecked()
        )

    def get_stop_server_autostop_timer_text(self):
        """
        Return the string to put on the stop server button, if there's an auto-stop timer
        """
        return strings._("gui_share_stop_server_autostop_timer")

    def autostop_timer_finished_should_stop_server(self):
        """
        The auto-stop timer expired, should we stop the server? Returns a bool
        """
        # If there were no attempts to download the share, or all downloads are done, we can stop
        if self.history.in_progress_count == 0 or self.web.done:
            self.server_status.stop_server()
            self.server_status_label.setText(strings._("close_on_autostop_timer"))
            return True
        # A download is probably still running - hold off on stopping the share
        else:
            self.server_status_label.setText(
                strings._("gui_share_mode_autostop_timer_waiting")
            )
            return False

    def start_server_custom(self):
        """
        Starting the server.
        """
        # Reset web counters
        self.web.share_mode.cur_history_id = 0

        # Hide and reset the downloads if we have previously shared
        self.reset_info_counters()

        self.remove_all_button.hide()

    def start_server_step2_custom(self):
        """
        Step 2 in starting the server. Zipping up files.
        """
        # Add progress bar to the status bar, indicating the compressing of files.
        self._zip_progress_bar = ZipProgressBar(self.common, 0)
        self.filenames = self.file_selection.get_filenames()

        self._zip_progress_bar.total_files_size = ShareMode._compute_total_size(
            self.filenames
        )
        self.status_bar.insertWidget(0, self._zip_progress_bar)

        # prepare the files for sending in a new thread
        self.compress_thread = CompressThread(self)
        self.compress_thread.success.connect(self.starting_server_step3.emit)
        self.compress_thread.success.connect(self.start_server_finished.emit)
        self.compress_thread.error.connect(self.starting_server_error.emit)
        self.server_status.server_canceled.connect(self.compress_thread.cancel)
        self.compress_thread.start()

    def start_server_step3_custom(self):
        """
        Step 3 in starting the server. Remove zip progress bar, and display large filesize
        warning, if applicable.
        """
        # Remove zip progress bar
        if self._zip_progress_bar is not None:
            self.status_bar.removeWidget(self._zip_progress_bar)
            self._zip_progress_bar = None

        # Warn about sending large files over Tor
        if self.web.share_mode.download_filesize >= 157286400:  # 150mb
            self.filesize_warning.setText(strings._("large_filesize"))
            self.filesize_warning.show()

    def start_server_error_custom(self):
        """
        Start server error.
        """
        if self._zip_progress_bar is not None:
            self.status_bar.removeWidget(self._zip_progress_bar)
            self._zip_progress_bar = None

    def stop_server_custom(self):
        """
        Stop server.
        """
        # Remove the progress bar
        if self._zip_progress_bar is not None:
            self.status_bar.removeWidget(self._zip_progress_bar)
            self._zip_progress_bar = None

        self.filesize_warning.hide()
        self.history.in_progress_count = 0
        self.history.completed_count = 0
        self.history.update_in_progress()
        self.file_selection.file_list.adjustSize()

        self.remove_all_button.show()

    def cancel_server_custom(self):
        """
        Stop the compression thread on cancel
        """
        if self.compress_thread:
            self.common.log("ShareMode", "cancel_server: quitting compress thread")
            self.compress_thread.quit()

    def handle_tor_broke_custom(self):
        """
        Connection to Tor broke.
        """
        self.primary_action.hide()

    def handle_request_started(self, event):
        """
        Handle REQUEST_STARTED event.
        """
        if event["data"]["use_gzip"]:
            filesize = self.web.share_mode.gzip_filesize
        else:
            filesize = self.web.share_mode.download_filesize

        item = ShareHistoryItem(self.common, event["data"]["id"], filesize)
        self.history.add(event["data"]["id"], item)
        self.toggle_history.update_indicator(True)
        self.history.in_progress_count += 1
        self.history.update_in_progress()

        self.system_tray.showMessage(
            strings._("systray_share_started_title"),
            strings._("systray_share_started_message"),
        )

    def handle_request_progress(self, event):
        """
        Handle REQUEST_PROGRESS event.
        """
        self.history.update(event["data"]["id"], event["data"]["bytes"])

        # Is the download complete?
        if event["data"]["bytes"] == self.web.share_mode.filesize:
            self.system_tray.showMessage(
                strings._("systray_share_completed_title"),
                strings._("systray_share_completed_message"),
            )

            # Update completed and in progress labels
            self.history.completed_count += 1
            self.history.in_progress_count -= 1
            self.history.update_completed()
            self.history.update_in_progress()

            # Close on finish?
            if self.settings.get("share", "autostop_sharing"):
                self.server_status.stop_server()
                self.status_bar.clearMessage()
                self.server_status_label.setText(strings._("closing_automatically"))
        else:
            if self.server_status.status == self.server_status.STATUS_STOPPED:
                self.history.cancel(event["data"]["id"])
                self.history.in_progress_count = 0
                self.history.update_in_progress()

    def handle_request_canceled(self, event):
        """
        Handle REQUEST_CANCELED event.
        """
        self.history.cancel(event["data"]["id"])

        # Update in progress count
        self.history.in_progress_count -= 1
        self.history.update_in_progress()
        self.system_tray.showMessage(
            strings._("systray_share_canceled_title"),
            strings._("systray_share_canceled_message"),
        )

    def update_primary_action(self):
        self.common.log("ShareMode", "update_primary_action")

        # Show or hide primary action layout
        file_count = self.file_selection.file_list.count()
        if file_count > 0:
            self.primary_action.show()
            self.info_label.show()
            self.remove_all_button.show()

            # Update the file count in the info label
            total_size_bytes = 0
            for index in range(self.file_selection.file_list.count()):
                item = self.file_selection.file_list.item(index)
                total_size_bytes += item.size_bytes
            total_size_readable = self.common.human_readable_filesize(total_size_bytes)

            if file_count > 1:
                self.info_label.setText(
                    strings._("gui_file_info").format(file_count, total_size_readable)
                )
            else:
                self.info_label.setText(
                    strings._("gui_file_info_single").format(
                        file_count, total_size_readable
                    )
                )

        else:
            self.primary_action.hide()
            self.info_label.hide()
            self.remove_all_button.hide()

    def reset_info_counters(self):
        """
        Set the info counters back to zero.
        """
        self.history.reset()
        self.toggle_history.indicator_count = 0
        self.toggle_history.update_indicator()

    def delete_all(self):
        """
        Delete All button clicked
        """
        self.file_selection.file_list.clear()
        self.file_selection.file_list.files_updated.emit()

        self.file_selection.file_list.setCurrentItem(None)

    @staticmethod
    def _compute_total_size(filenames):
        total_size = 0
        for filename in filenames:
            if os.path.isfile(filename):
                total_size += os.path.getsize(filename)
            if os.path.isdir(filename):
                total_size += Common.dir_size(filename)
        return total_size


class ZipProgressBar(QtWidgets.QProgressBar):
    update_processed_size_signal = QtCore.Signal(int)

    def __init__(self, common, total_files_size):
        super(ZipProgressBar, self).__init__()
        self.common = common

        self.setMaximumHeight(20)
        self.setMinimumWidth(200)
        self.setValue(0)
        self.setFormat(strings._("zip_progress_bar_format"))
        self.setStyleSheet(self.common.gui.css["share_zip_progess_bar"])

        self._total_files_size = total_files_size
        self._processed_size = 0

        self.update_processed_size_signal.connect(self.update_processed_size)

    @property
    def total_files_size(self):
        return self._total_files_size

    @total_files_size.setter
    def total_files_size(self, val):
        self._total_files_size = val

    @property
    def processed_size(self):
        return self._processed_size

    @processed_size.setter
    def processed_size(self, val):
        self.update_processed_size(val)

    def update_processed_size(self, val):
        self._processed_size = val

        if self.processed_size < self.total_files_size:
            self.setValue(int((self.processed_size * 100) / self.total_files_size))
        elif self.total_files_size != 0:
            self.setValue(100)
        else:
            self.setValue(0)



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import textwrap
from PySide6 import QtCore, QtWidgets, QtGui
from PySide6.QtCore import Qt

from .. import strings
from ..widgets import Alert
from ..widgets import QRCodeDialog
from ..gui_common import GuiCommon


class ServerStatus(QtWidgets.QWidget):
    """
    The server status chunk of the GUI.
    """

    server_started = QtCore.Signal()
    server_started_finished = QtCore.Signal()
    server_stopped = QtCore.Signal()
    server_canceled = QtCore.Signal()
    button_clicked = QtCore.Signal()
    url_copied = QtCore.Signal()
    client_auth_copied = QtCore.Signal()

    STATUS_STOPPED = 0
    STATUS_WORKING = 1
    STATUS_STARTED = 2

    def __init__(
        self,
        common,
        qtapp,
        app,
        mode_settings,
        mode_settings_widget,
        file_selection=None,
        local_only=False,
    ):
        super(ServerStatus, self).__init__()

        self.common = common

        self.status = self.STATUS_STOPPED
        self.mode = None  # Gets set in self.set_mode

        self.qtapp = qtapp
        self.app = app
        self.settings = mode_settings
        self.mode_settings_widget = mode_settings_widget

        self.web = None
        self.autostart_timer_datetime = None
        self.local_only = local_only

        self.resizeEvent(None)

        # Server layout
        self.server_button = QtWidgets.QPushButton()
        self.server_button.clicked.connect(self.server_button_clicked)

        # URL layout
        url_font = QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont)
        self.url_description = QtWidgets.QLabel()
        self.url_description.setWordWrap(True)
        self.url_description.setMinimumHeight(50)

        # URL sharing instructions, above the URL and Copy Address/QR Code buttons
        self.url_instructions = QtWidgets.QLabel()
        self.url_instructions.setWordWrap(True)

        # The URL label itself
        self.url = QtWidgets.QLabel()
        self.url.setFont(url_font)
        self.url.setWordWrap(True)
        self.url.setMinimumSize(self.url.sizeHint())
        self.url.setStyleSheet(self.common.gui.css["server_status_url"])
        self.url.setTextInteractionFlags(
            Qt.TextSelectableByMouse | Qt.TextSelectableByKeyboard
        )

        # Copy Onion Address button
        self.copy_url_button = QtWidgets.QPushButton(strings._("gui_copy_url"))
        self.copy_url_button.setStyleSheet(
            self.common.gui.css["server_status_url_buttons"]
        )
        self.copy_url_button.clicked.connect(self.copy_url)

        # Onion Address QR code button
        self.show_url_qr_code_button = QtWidgets.QPushButton(
            strings._("gui_show_qr_code")
        )
        self.show_url_qr_code_button.hide()
        self.show_url_qr_code_button.clicked.connect(
            self.show_url_qr_code_button_clicked
        )
        self.show_url_qr_code_button.setStyleSheet(
            self.common.gui.css["server_status_url_buttons"]
        )

        # Client Auth sharing instructions, above the
        # Copy Private Key/QR Code buttons
        self.client_auth_instructions = QtWidgets.QLabel()
        self.client_auth_instructions.setWordWrap(True)
        self.client_auth_instructions.setText(strings._("gui_client_auth_instructions"))

        # The private key itself
        self.private_key = QtWidgets.QLabel()
        self.private_key.setFont(url_font)
        self.private_key.setWordWrap(True)
        self.private_key.setMinimumSize(self.private_key.sizeHint())
        self.private_key.setStyleSheet(self.common.gui.css["server_status_url"])
        self.private_key.setTextInteractionFlags(Qt.NoTextInteraction)
        self.private_key_hidden = True

        # Copy ClientAuth button
        self.copy_client_auth_button = QtWidgets.QPushButton(
            strings._("gui_copy_client_auth")
        )
        self.copy_client_auth_button.setStyleSheet(
            self.common.gui.css["server_status_url_buttons"]
        )
        self.copy_client_auth_button.clicked.connect(self.copy_client_auth)

        # ClientAuth QR code button
        self.show_client_auth_qr_code_button = QtWidgets.QPushButton(
            strings._("gui_show_qr_code")
        )
        self.show_client_auth_qr_code_button.hide()
        self.show_client_auth_qr_code_button.clicked.connect(
            self.show_client_auth_qr_code_button_clicked
        )
        self.show_client_auth_qr_code_button.setStyleSheet(
            self.common.gui.css["server_status_url_buttons"]
        )

        # ClientAuth reveal/hide toggle button
        self.client_auth_toggle_button = QtWidgets.QPushButton(strings._("gui_reveal"))
        self.client_auth_toggle_button.hide()
        self.client_auth_toggle_button.clicked.connect(
            self.client_auth_toggle_button_clicked
        )
        self.client_auth_toggle_button.setStyleSheet(
            self.common.gui.css["server_status_url_buttons"]
        )

        # URL instructions layout
        url_buttons_layout = QtWidgets.QHBoxLayout()
        url_buttons_layout.addWidget(self.copy_url_button)
        url_buttons_layout.addWidget(self.show_url_qr_code_button)
        url_buttons_layout.addStretch()

        url_layout = QtWidgets.QVBoxLayout()
        url_layout.addWidget(self.url_description)
        url_layout.addWidget(self.url_instructions)
        url_layout.addWidget(self.url)
        url_layout.addLayout(url_buttons_layout)

        # Private key instructions layout
        client_auth_buttons_layout = QtWidgets.QHBoxLayout()
        client_auth_buttons_layout.addWidget(self.copy_client_auth_button)
        client_auth_buttons_layout.addWidget(self.show_client_auth_qr_code_button)
        client_auth_buttons_layout.addWidget(self.client_auth_toggle_button)
        client_auth_buttons_layout.addStretch()

        client_auth_layout = QtWidgets.QVBoxLayout()
        client_auth_layout.addWidget(self.client_auth_instructions)
        client_auth_layout.addWidget(self.private_key)
        client_auth_layout.addLayout(client_auth_buttons_layout)

        # Add the widgets and URL/ClientAuth layouts
        # to the main ServerStatus layout
        button_layout = QtWidgets.QHBoxLayout()
        button_layout.addWidget(self.server_button)
        button_layout.addStretch()

        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(button_layout)
        layout.addLayout(url_layout)
        layout.addLayout(client_auth_layout)
        self.setLayout(layout)

    def set_mode(self, share_mode, file_selection=None):
        """
        The server status is in share mode.
        """
        self.mode = share_mode

        if (self.mode == self.common.gui.MODE_SHARE) or (
            self.mode == self.common.gui.MODE_WEBSITE
        ):
            self.file_selection = file_selection

        self.update()

    def resizeEvent(self, event):
        """
        When the widget is resized, try and adjust the display of a v3 onion URL.
        """
        try:
            # Wrap the URL label
            url_length = len(self.get_url())
            if url_length > 60:
                width = self.frameGeometry().width()
                if width < 530:
                    wrapped_onion_url = textwrap.fill(self.get_url(), 46)
                    self.url.setText(wrapped_onion_url)
                else:
                    self.url.setText(self.get_url())
        except Exception:
            pass

    def show_url(self):
        """
        Show the URL in the UI.
        """
        self.url_description.show()

        info_image = GuiCommon.get_resource_path("images/info.png")

        if self.mode == self.common.gui.MODE_SHARE:
            if self.settings.get("general", "public"):
                self.url_description.setText(
                    strings._("gui_share_url_public_description").format(info_image)
                )
            else:
                self.url_description.setText(
                    strings._("gui_share_url_description").format(info_image)
                )
        elif self.mode == self.common.gui.MODE_WEBSITE:
            if self.settings.get("general", "public"):
                self.url_description.setText(
                    strings._("gui_website_url_public_description").format(info_image)
                )
            else:
                self.url_description.setText(
                    strings._("gui_website_url_description").format(info_image)
                )
        elif self.mode == self.common.gui.MODE_RECEIVE:
            if self.settings.get("general", "public"):
                self.url_description.setText(
                    strings._("gui_receive_url_public_description").format(info_image)
                )
            else:
                self.url_description.setText(
                    strings._("gui_receive_url_description").format(info_image)
                )
        elif self.mode == self.common.gui.MODE_CHAT:
            if self.settings.get("general", "public"):
                self.url_description.setText(
                    strings._("gui_chat_url_public_description").format(info_image)
                )
            else:
                self.url_description.setText(
                    strings._("gui_chat_url_description").format(info_image)
                )

        # Show a Tool Tip explaining the lifecycle of this URL
        if self.settings.get("persistent", "enabled"):
            if self.mode == self.common.gui.MODE_SHARE and self.settings.get(
                "share", "autostop_sharing"
            ):
                self.url_description.setToolTip(
                    strings._("gui_url_label_onetime_and_persistent")
                )
            else:
                self.url_description.setToolTip(strings._("gui_url_label_persistent"))
        else:
            if self.mode == self.common.gui.MODE_SHARE and self.settings.get(
                "share", "autostop_sharing"
            ):
                self.url_description.setToolTip(strings._("gui_url_label_onetime"))
            else:
                self.url_description.setToolTip(strings._("gui_url_label_stay_open"))

        if self.settings.get("general", "public"):
            self.url_instructions.setText(strings._("gui_url_instructions_public_mode"))
        else:
            self.url_instructions.setText(strings._("gui_url_instructions"))
        self.url_instructions.show()
        self.url.setText(self.get_url())
        self.url.show()
        self.copy_url_button.show()
        self.show_url_qr_code_button.show()

        if self.settings.get("general", "public"):
            self.client_auth_instructions.hide()
            self.private_key.hide()
            self.copy_client_auth_button.hide()
            self.show_client_auth_qr_code_button.hide()
        else:
            self.client_auth_instructions.show()
            if self.private_key_hidden:
                self.private_key.setText("*" * len(self.app.auth_string))
                self.private_key.setTextInteractionFlags(Qt.NoTextInteraction)
            else:
                self.private_key.setText(self.app.auth_string)
                self.private_key.setTextInteractionFlags(
                    Qt.TextSelectableByMouse | Qt.TextSelectableByKeyboard
                )
            self.private_key.show()
            self.copy_client_auth_button.show()
            self.show_client_auth_qr_code_button.show()
            self.client_auth_toggle_button.show()

    def update(self):
        """
        Update the GUI elements based on the current state.
        """
        self.common.log("ServerStatus", "update")
        # Set the URL fields
        if self.status == self.STATUS_STARTED:
            # The backend Onion may have saved new settings, such as the private key.
            # Reload the settings before saving new ones.
            self.common.settings.load()
            self.show_url()

            if self.settings.get("general", "autostop_timer"):
                self.server_button.setToolTip(
                    strings._("gui_stop_server_autostop_timer_tooltip").format(
                        self.mode_settings_widget.autostop_timer_widget.dateTime().toString(
                            "h:mm AP, MMMM dd, yyyy"
                        )
                    )
                )
        else:
            self.url_description.hide()
            self.url_instructions.hide()
            self.url.hide()
            self.copy_url_button.hide()
            self.show_url_qr_code_button.hide()
            self.private_key.hide()
            self.client_auth_instructions.hide()
            self.copy_client_auth_button.hide()
            self.show_client_auth_qr_code_button.hide()
            self.client_auth_toggle_button.hide()

            self.mode_settings_widget.update_ui()

        # Button
        if (
            self.mode == self.common.gui.MODE_SHARE
            and self.file_selection.get_num_files() == 0
        ):
            self.server_button.hide()
        elif (
            self.mode == self.common.gui.MODE_WEBSITE
            and self.file_selection.get_num_files() == 0
        ):
            self.server_button.hide()
        else:
            self.server_button.show()

            if self.status == self.STATUS_STOPPED:
                self.server_button.setStyleSheet(
                    self.common.gui.css["server_status_button_stopped"]
                )
                self.server_button.setEnabled(True)
                if self.mode == self.common.gui.MODE_SHARE:
                    self.server_button.setText(strings._("gui_share_start_server"))
                elif self.mode == self.common.gui.MODE_WEBSITE:
                    self.server_button.setText(strings._("gui_share_start_server"))
                elif self.mode == self.common.gui.MODE_CHAT:
                    self.server_button.setText(strings._("gui_chat_start_server"))
                else:
                    self.server_button.setText(strings._("gui_receive_start_server"))
                self.server_button.setToolTip("")
            elif self.status == self.STATUS_STARTED:
                self.server_button.setStyleSheet(
                    self.common.gui.css["server_status_button_started"]
                )
                self.server_button.setEnabled(True)
                if self.mode == self.common.gui.MODE_SHARE:
                    self.server_button.setText(strings._("gui_share_stop_server"))
                elif self.mode == self.common.gui.MODE_WEBSITE:
                    self.server_button.setText(strings._("gui_share_stop_server"))
                elif self.mode == self.common.gui.MODE_CHAT:
                    self.server_button.setText(strings._("gui_chat_stop_server"))
                else:
                    self.server_button.setText(strings._("gui_receive_stop_server"))
            elif self.status == self.STATUS_WORKING:
                self.server_button.setStyleSheet(
                    self.common.gui.css["server_status_button_working"]
                )
                self.server_button.setEnabled(True)
                if self.settings.get("general", "autostart_timer"):
                    self.server_button.setToolTip(
                        strings._("gui_start_server_autostart_timer_tooltip").format(
                            self.mode_settings_widget.autostart_timer_widget.dateTime().toString(
                                "h:mm AP, MMMM dd, yyyy"
                            )
                        )
                    )
                else:
                    if self.common.platform == "Windows":
                        self.server_button.setText(strings._("gui_please_wait"))
                    else:
                        self.server_button.setText(
                            strings._("gui_please_wait_no_button")
                        )
            else:
                self.server_button.setStyleSheet(
                    self.common.gui.css["server_status_button_working"]
                )
                self.server_button.setEnabled(False)
                self.server_button.setText(strings._("gui_please_wait_no_button"))

    def server_button_clicked(self):
        """
        Toggle starting or stopping the server.
        """
        if self.status == self.STATUS_STOPPED:
            can_start = True
            if self.settings.get("general", "autostart_timer"):
                if self.local_only:
                    self.autostart_timer_datetime = (
                        self.mode_settings_widget.autostart_timer_widget.dateTime().toPython()
                    )
                else:
                    self.autostart_timer_datetime = (
                        self.mode_settings_widget.autostart_timer_widget.dateTime()
                        .toPython()
                        .replace(second=0, microsecond=0)
                    )
                # If the timer has actually passed already before the user hit Start, refuse to start the server.
                if (
                    QtCore.QDateTime.currentDateTime().toPython()
                    > self.autostart_timer_datetime
                ):
                    can_start = False
                    Alert(
                        self.common,
                        strings._("gui_server_autostart_timer_expired"),
                        QtWidgets.QMessageBox.Warning,
                    )
            if self.settings.get("general", "autostop_timer"):
                if self.local_only:
                    self.autostop_timer_datetime = (
                        self.mode_settings_widget.autostop_timer_widget.dateTime().toPython()
                    )
                else:
                    # Get the timer chosen, stripped of its seconds. This prevents confusion if the share stops at (say) 37 seconds past the minute chosen
                    self.autostop_timer_datetime = (
                        self.mode_settings_widget.autostop_timer_widget.dateTime()
                        .toPython()
                        .replace(second=0, microsecond=0)
                    )
                # If the timer has actually passed already before the user hit Start, refuse to start the server.
                if (
                    QtCore.QDateTime.currentDateTime().toPython()
                    > self.autostop_timer_datetime
                ):
                    can_start = False
                    Alert(
                        self.common,
                        strings._("gui_server_autostop_timer_expired"),
                        QtWidgets.QMessageBox.Warning,
                    )
                if self.settings.get("general", "autostart_timer"):
                    if self.autostop_timer_datetime <= self.autostart_timer_datetime:
                        Alert(
                            self.common,
                            strings._(
                                "gui_autostop_timer_cant_be_earlier_than_autostart_timer"
                            ),
                            QtWidgets.QMessageBox.Warning,
                        )
                        can_start = False
            if can_start:
                self.start_server()
        elif self.status == self.STATUS_STARTED:
            self.stop_server()
        elif self.status == self.STATUS_WORKING and (
            self.common.platform == "Windows"
            or self.settings.get("general", "autostart_timer")
        ):
            self.cancel_server()
        self.button_clicked.emit()

    def show_url_qr_code_button_clicked(self):
        """
        Show a QR code of the onion URL.
        """
        self.qr_code_dialog = QRCodeDialog(
            self.common, strings._("gui_qr_label_url_title"), self.get_url()
        )

    def show_client_auth_qr_code_button_clicked(self):
        """
        Show a QR code of the private key
        """
        self.qr_code_dialog = QRCodeDialog(
            self.common,
            strings._("gui_qr_label_auth_string_title"),
            self.app.auth_string,
        )

    def client_auth_toggle_button_clicked(self):
        """
        ClientAuth reveal/hide toggle button clicked
        """
        if self.private_key_hidden:
            self.private_key_hidden = False
            self.client_auth_toggle_button.setText(strings._("gui_hide"))
        else:
            self.private_key_hidden = True
            self.client_auth_toggle_button.setText(strings._("gui_reveal"))

        self.show_url()

    def start_server(self):
        """
        Start the server.
        """
        self.status = self.STATUS_WORKING
        self.update()
        self.server_started.emit()

    def start_server_finished(self):
        """
        The server has finished starting.
        """
        self.status = self.STATUS_STARTED
        # self.copy_url()
        self.update()
        self.server_started_finished.emit()

    def stop_server(self):
        """
        Stop the server.
        """
        self.status = self.STATUS_WORKING
        self.mode_settings_widget.autostart_timer_reset()
        self.mode_settings_widget.autostop_timer_reset()
        self.update()
        self.server_stopped.emit()

    def cancel_server(self):
        """
        Cancel the server.
        """
        self.common.log(
            "ServerStatus", "cancel_server", "Canceling the server mid-startup"
        )
        self.status = self.STATUS_WORKING
        self.mode_settings_widget.autostart_timer_reset()
        self.mode_settings_widget.autostop_timer_reset()
        self.update()
        self.server_canceled.emit()

    def stop_server_finished(self):
        """
        The server has finished stopping.
        """
        self.status = self.STATUS_STOPPED
        self.update()

    def copy_url(self):
        """
        Copy the onionshare URL to the clipboard.
        """
        clipboard = self.qtapp.clipboard()
        clipboard.setText(self.get_url())

        self.url_copied.emit()

    def copy_client_auth(self):
        """
        Copy the ClientAuth private key line to the clipboard.
        """
        clipboard = self.qtapp.clipboard()
        clipboard.setText(self.app.auth_string)

        self.client_auth_copied.emit()

    def get_url(self):
        """
        Returns the OnionShare URL.
        """
        url = f"http://{self.app.onion_host}"
        return url



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from PySide6 import QtCore, QtWidgets

from onionshare_cli.common import AutoStopTimer

from .history import IndividualFileHistoryItem
from .mode_settings_widget import ModeSettingsWidget

from ..server_status import ServerStatus
from ... import strings
from ...threads import OnionThread, WebThread, AutoStartTimer
from ...widgets import Alert, MinimumSizeWidget


class Mode(QtWidgets.QWidget):
    """
    The class that all modes inherit from
    """

    start_server_finished = QtCore.Signal()
    stop_server_finished = QtCore.Signal()
    starting_server_step2 = QtCore.Signal()
    starting_server_step3 = QtCore.Signal()
    starting_server_error = QtCore.Signal(str)
    starting_server_early = QtCore.Signal()
    set_server_active = QtCore.Signal(bool)
    change_persistent = QtCore.Signal(int, bool)

    def __init__(self, tab):
        super(Mode, self).__init__()
        self.tab = tab
        self.settings = tab.settings

        self.common = tab.common
        self.qtapp = self.common.gui.qtapp
        self.app = tab.app

        self.status_bar = tab.status_bar
        self.server_status_label = tab.status_bar.server_status_label
        self.system_tray = tab.system_tray

        self.filenames = tab.filenames

        # The web object gets created in init()
        self.web = None

        # Threads start out as None
        self.onion_thread = None
        self.web_thread = None
        self.startup_thread = None

        # Mode settings widget
        self.mode_settings_widget = ModeSettingsWidget(
            self.common, self.tab, self.settings
        )
        self.mode_settings_widget.change_persistent.connect(self.change_persistent)

        # Server status
        self.server_status = ServerStatus(
            self.common,
            self.qtapp,
            self.app,
            self.settings,
            self.mode_settings_widget,
            None,
            self.common.gui.local_only,
        )
        self.server_status.server_started.connect(self.start_server)
        self.server_status.server_stopped.connect(self.stop_server)
        self.server_status.server_canceled.connect(self.cancel_server)
        self.start_server_finished.connect(self.server_status.start_server_finished)
        self.stop_server_finished.connect(self.server_status.stop_server_finished)
        self.starting_server_step2.connect(self.start_server_step2)
        self.starting_server_step3.connect(self.start_server_step3)
        self.starting_server_early.connect(self.start_server_early)
        self.starting_server_error.connect(self.start_server_error)

        # Primary action
        # Note: It's up to the downstream Mode to add this to its layout
        self.primary_action_layout = QtWidgets.QVBoxLayout()
        self.primary_action_layout.addWidget(self.mode_settings_widget)
        self.primary_action = QtWidgets.QWidget()
        self.primary_action.setLayout(self.primary_action_layout)

        # It's up to the downstream Mode to add stuff to self.content_layout
        # self.content_layout shows the actual content of the mode
        # self.tor_not_connected_layout is displayed when Tor isn't connected
        self.content_layout = QtWidgets.QVBoxLayout()
        self.content_widget = QtWidgets.QWidget()
        self.content_widget.setLayout(self.content_layout)

        tor_not_connected_label = QtWidgets.QLabel(
            strings._("mode_tor_not_connected_label")
        )
        tor_not_connected_label.setAlignment(QtCore.Qt.AlignHCenter)
        tor_not_connected_label.setStyleSheet(
            self.common.gui.css["tor_not_connected_label"]
        )
        self.tor_not_connected_layout = QtWidgets.QVBoxLayout()
        self.tor_not_connected_layout.addStretch()
        self.tor_not_connected_layout.addWidget(tor_not_connected_label)
        self.tor_not_connected_layout.addWidget(MinimumSizeWidget(700, 0))
        self.tor_not_connected_layout.addStretch()
        self.tor_not_connected_widget = QtWidgets.QWidget()
        self.tor_not_connected_widget.setLayout(self.tor_not_connected_layout)

        self.wrapper_layout = QtWidgets.QVBoxLayout()
        self.wrapper_layout.addWidget(self.content_widget)
        self.wrapper_layout.addWidget(self.tor_not_connected_widget)
        self.setLayout(self.wrapper_layout)

        if self.common.gui.onion.connected_to_tor:
            self.tor_connection_started()
        else:
            self.tor_connection_stopped()

    def init(self):
        """
        Add custom initialization here.
        """
        pass

    def get_type(self):
        """
        Returns the type of mode as a string (e.g. "share", "receive", etc.)
        """
        pass

    def human_friendly_time(self, secs):
        """
        Returns a human-friendly time delta from given seconds.
        """
        days = secs // 86400
        hours = (secs - days * 86400) // 3600
        minutes = (secs - days * 86400 - hours * 3600) // 60
        seconds = secs - days * 86400 - hours * 3600 - minutes * 60
        if not seconds:
            seconds = "0"
        result = (
            (f"{days}{strings._('days_first_letter')}, " if days else "")
            + (f"{hours}{strings._('hours_first_letter')}, " if hours else "")
            + (f"{minutes}{strings._('minutes_first_letter')}, " if minutes else "")
            + f"{seconds}{strings._('seconds_first_letter')}"
        )

        return result

    def timer_callback(self):
        """
        This method is called regularly on a timer.
        """
        # If this is a scheduled share, display the countdown til the share starts
        if self.server_status.status == ServerStatus.STATUS_WORKING:
            if self.settings.get("general", "autostart_timer"):
                now = QtCore.QDateTime.currentDateTime()
                if self.server_status.local_only:
                    seconds_remaining = now.secsTo(
                        self.mode_settings_widget.autostart_timer_widget.dateTime()
                    )
                else:
                    seconds_remaining = now.secsTo(
                        self.server_status.autostart_timer_datetime.replace(
                            second=0, microsecond=0
                        )
                    )
                # Update the server button
                if seconds_remaining > 0:
                    self.server_status.server_button.setText(
                        strings._("gui_waiting_to_start").format(
                            self.human_friendly_time(seconds_remaining)
                        )
                    )
                else:
                    if self.common.platform == "Windows" or self.settings.get(
                        "general", "autostart_timer"
                    ):
                        self.server_status.server_button.setText(
                            strings._("gui_please_wait")
                        )
                    else:
                        self.server_status.server_button.setText(
                            strings._("gui_please_wait_no_button")
                        )

        # If the auto-stop timer has stopped, stop the server
        if self.server_status.status == ServerStatus.STATUS_STARTED:
            if self.app.autostop_timer_thread and self.settings.get(
                "general", "autostop_timer"
            ):
                if self.autostop_timer_datetime_delta > 0:
                    now = QtCore.QDateTime.currentDateTime()
                    seconds_remaining = now.secsTo(
                        self.server_status.autostop_timer_datetime
                    )

                    # Update the server button
                    server_button_text = self.get_stop_server_autostop_timer_text()
                    self.server_status.server_button.setText(
                        server_button_text.format(
                            self.human_friendly_time(seconds_remaining)
                        )
                    )

                    self.status_bar.clearMessage()
                    if not self.app.autostop_timer_thread.is_alive():
                        self.autostop_timer_finished_should_stop_server()

    def timer_callback_custom(self):
        """
        Add custom timer code.
        """
        pass

    def get_stop_server_autostop_timer_text(self):
        """
        Return the string to put on the stop server button, if there's an auto-stop timer
        """
        pass

    def autostop_timer_finished_should_stop_server(self):
        """
        The auto-stop timer expired, should we stop the server? Returns a bool
        """
        pass

    def start_server(self):
        """
        Start the onionshare server. This uses multiple threads to start the Tor onion
        server and the web app.
        """
        self.common.log("Mode", "start_server")

        self.start_server_custom()
        self.set_server_active.emit(True)

        # Clear the status bar
        self.status_bar.clearMessage()
        self.server_status_label.setText("")

        # Hide the mode settings
        self.mode_settings_widget.hide()

        # Ensure we always get a new random port each time we might launch an OnionThread
        self.app.port = None

        # Start the onion thread. If this share was scheduled for a future date,
        # the OnionThread will start and exit 'early' to obtain the port, password
        # and onion address, but it will not start the WebThread yet.
        if self.settings.get("general", "autostart_timer"):
            self.start_onion_thread(obtain_onion_early=True)
            self.common.log("Mode", "start_server", "Starting auto-start timer")
            self.startup_thread = AutoStartTimer(self)
            # Once the timer has finished, start the real share, with a WebThread
            self.startup_thread.success.connect(self.start_scheduled_service)
            self.startup_thread.error.connect(self.start_server_error)
            self.startup_thread.canceled = False
            self.startup_thread.start()
        else:
            self.start_onion_thread()

    def start_onion_thread(self, obtain_onion_early=False):
        # If we tried to start with Client Auth and our Tor is too old to support it,
        # bail out early
        if (
            not self.server_status.local_only
            and not self.app.onion.supports_stealth
            and not self.settings.get("general", "public")
        ):
            self.stop_server()
            self.start_server_error(strings._("gui_server_doesnt_support_stealth"))
        else:
            self.common.log("Mode", "start_server", "Starting an onion thread")
            self.obtain_onion_early = obtain_onion_early
            self.onion_thread = OnionThread(self)
            self.onion_thread.success.connect(self.starting_server_step2.emit)
            self.onion_thread.success_early.connect(self.starting_server_early.emit)
            self.onion_thread.error.connect(self.starting_server_error.emit)
            self.web_thread = WebThread(self)
            self.web_thread.error.connect(self.starting_server_error.emit)
            self.onion_thread.start()

    def start_scheduled_service(self, obtain_onion_early=False):
        # We start a new OnionThread with the saved scheduled key from settings
        self.common.settings.load()
        self.obtain_onion_early = obtain_onion_early
        self.common.log("Mode", "start_server", "Starting a scheduled onion thread")
        self.onion_thread = OnionThread(self)
        self.onion_thread.success.connect(self.starting_server_step2.emit)
        self.onion_thread.error.connect(self.starting_server_error.emit)
        self.onion_thread.start()

    def start_server_custom(self):
        """
        Add custom initialization here.
        """
        pass

    def start_server_early(self):
        """
        An 'early' start of an onion service in order to obtain the onion
        address for a scheduled start. Shows the onion address in the UI
        in advance of actually starting the share.
        """
        self.server_status.show_url()

    def start_server_step2(self):
        """
        Step 2 in starting the onionshare server.
        """
        self.common.log("Mode", "start_server_step2")

        self.start_server_step2_custom()

        # Nothing to do here.

        # start_server_step2_custom has call these to move on:
        # self.starting_server_step3.emit()
        # self.start_server_finished.emit()

    def start_server_step2_custom(self):
        """
        Add custom initialization here.
        """
        pass

    def start_server_step3(self):
        """
        Step 3 in starting the onionshare server.
        """
        self.common.log("Mode", "start_server_step3")

        self.start_server_step3_custom()

        if self.settings.get("general", "autostop_timer"):
            # Convert the date value to seconds between now and then
            now = QtCore.QDateTime.currentDateTime()
            self.autostop_timer_datetime_delta = now.secsTo(
                self.server_status.autostop_timer_datetime
            )
            # Start the auto-stop timer
            if self.autostop_timer_datetime_delta > 0:
                self.app.autostop_timer_thread = AutoStopTimer(
                    self.common, self.autostop_timer_datetime_delta
                )
                self.app.autostop_timer_thread.start()
            # The auto-stop timer has actually already passed since the user clicked Start. Probably the Onion service took too long to start.
            else:
                self.stop_server()
                self.start_server_error(
                    strings._("gui_server_started_after_autostop_timer")
                )

    def start_server_step3_custom(self):
        """
        Add custom initialization here.
        """
        pass

    def start_server_error(self, error):
        """
        If there's an error when trying to start the onion service
        """
        self.common.log("Mode", "start_server_error")

        Alert(self.common, error, QtWidgets.QMessageBox.Warning)
        self.set_server_active.emit(False)
        self.server_status.stop_server()
        self.status_bar.clearMessage()

        self.start_server_error_custom()

    def start_server_error_custom(self):
        """
        Add custom initialization here.
        """
        pass

    def cancel_server(self):
        """
        Cancel the server while it is preparing to start
        """
        self.cancel_server_custom()
        if self.startup_thread:
            self.common.log("Mode", "cancel_server: quitting startup thread")
            self.startup_thread.canceled = True
            self.app.onion.scheduled_key = None
            self.app.onion.scheduled_auth_cookie = None
            self.startup_thread.quit()

        # Canceling only works in Windows
        # https://github.com/onionshare/onionshare/issues/1371
        if self.common.platform == "Windows":
            if self.onion_thread:
                self.common.log("Mode", "cancel_server: quitting onion thread")
                self.onion_thread.terminate()
                self.onion_thread.wait()
            if self.web_thread:
                self.common.log("Mode", "cancel_server: quitting web thread")
                self.web_thread.terminate()
                self.web_thread.wait()

        self.stop_server()

    def cancel_server_custom(self):
        """
        Add custom initialization here.
        """
        pass

    def stop_server(self):
        """
        Stop the onionshare server.
        """
        self.common.log("Mode", "stop_server")

        if self.server_status.status != ServerStatus.STATUS_STOPPED:
            try:
                self.web.stop(self.app.port)
            except Exception:
                # Probably we had no port to begin with (Onion service didn't start)
                pass
        self.web.cleanup()

        self.stop_server_custom()

        self.set_server_active.emit(False)
        self.stop_server_finished.emit()

        # Show the mode settings
        self.mode_settings_widget.show()

    def stop_server_custom(self):
        """
        Add custom initialization here.
        """
        pass

    def handle_tor_broke(self):
        """
        Handle connection from Tor breaking.
        """
        if self.server_status.status != ServerStatus.STATUS_STOPPED:
            self.server_status.stop_server()
        self.handle_tor_broke_custom()

    def handle_tor_broke_custom(self):
        """
        Add custom initialization here.
        """
        pass

    # Handle web server events

    def handle_request_load(self, event):
        """
        Handle REQUEST_LOAD event.
        """
        pass

    def handle_request_started(self, event):
        """
        Handle REQUEST_STARTED event.
        """
        pass

    def handle_request_progress(self, event):
        """
        Handle REQUEST_PROGRESS event.
        """
        pass

    def handle_request_canceled(self, event):
        """
        Handle REQUEST_CANCELED event.
        """
        pass

    def handle_request_upload_includes_message(self, event):
        """
        Handle REQUEST_UPLOAD_INCLUDES_MESSAGE event.
        """
        pass

    def handle_request_upload_file_renamed(self, event):
        """
        Handle REQUEST_UPLOAD_FILE_RENAMED event.
        """
        pass

    def handle_request_upload_message(self, event):
        """
        Handle REQUEST_UPLOAD_MESSAGE event.
        """
        pass

    def handle_request_upload_set_dir(self, event):
        """
        Handle REQUEST_UPLOAD_SET_DIR event.
        """
        pass

    def handle_request_upload_finished(self, event):
        """
        Handle REQUEST_UPLOAD_FINISHED event.
        """
        pass

    def handle_request_upload_canceled(self, event):
        """
        Handle REQUEST_UPLOAD_CANCELED event.
        """
        pass

    def handle_request_individual_file_started(self, event):
        """
        Handle REQUEST_INDVIDIDUAL_FILES_STARTED event.
        Used in both Share and Website modes, so implemented here.
        """
        self.toggle_history.update_indicator(True)
        self.history.requests_count += 1
        self.history.update_requests()

        item = IndividualFileHistoryItem(self.common, event["data"], event["path"])
        self.history.add(event["data"]["id"], item)

    def handle_request_individual_file_progress(self, event):
        """
        Handle REQUEST_INDVIDIDUAL_FILES_PROGRESS event.
        Used in both Share and Website modes, so implemented here.
        """
        self.history.update(event["data"]["id"], event["data"]["bytes"])

        if self.server_status.status == self.server_status.STATUS_STOPPED:
            self.history.cancel(event["data"]["id"])

    def handle_request_individual_file_canceled(self, event):
        """
        Handle REQUEST_INDVIDIDUAL_FILES_CANCELED event.
        Used in both Share and Website modes, so implemented here.
        """
        self.history.cancel(event["data"]["id"])

    def tor_connection_started(self):
        """
        This is called on every Mode when Tor is connected
        """
        self.content_widget.show()
        self.tor_not_connected_widget.hide()
        self.primary_action.show()
        if not self.tab.timer.isActive():
            self.tab.timer.start(500)
        if self.settings.get("persistent", "enabled") and self.settings.get("persistent", "autostart_on_launch"):
            self.server_status.start_server()

    def tor_connection_stopped(self):
        """
        This is called on every Mode when Tor is disconnected
        """
        if self.common.gui.local_only:
            self.tor_connection_started()
            return

        self.content_widget.hide()
        self.tor_not_connected_widget.show()



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
from PySide6 import QtCore, QtWidgets, QtGui

from ... import strings
from ...widgets import Alert, AddFileDialog
from ...gui_common import GuiCommon


class DropHereWidget(QtWidgets.QWidget):
    """
    When there are no files or folders in the FileList yet, display the
    'drop files here' message and graphic.
    """

    def __init__(self, common, image_filename, header_text, w, h, parent):
        super(DropHereWidget, self).__init__(parent)
        self.common = common
        self.setAcceptDrops(True)

        self.image_label = QtWidgets.QLabel(parent=self)
        self.image_label.setPixmap(
            QtGui.QPixmap.fromImage(
                QtGui.QImage(GuiCommon.get_resource_path(image_filename))
            )
        )
        self.image_label.setAlignment(QtCore.Qt.AlignCenter)
        self.image_label.show()

        self.header_label = QtWidgets.QLabel(parent=self)
        self.header_label.setText(header_text)
        self.header_label.setStyleSheet(
            self.common.gui.css["share_file_selection_drop_here_header_label"]
        )
        self.header_label.setAlignment(QtCore.Qt.AlignCenter)
        self.header_label.show()

        self.text_label = QtWidgets.QLabel(parent=self)
        self.text_label.setText(strings._("gui_drag_and_drop"))
        self.text_label.setStyleSheet(
            self.common.gui.css["share_file_selection_drop_here_label"]
        )
        self.text_label.setAlignment(QtCore.Qt.AlignCenter)
        self.text_label.show()

        self.resize(w, h)
        self.hide()

    def dragEnterEvent(self, event):
        self.hide()
        event.accept()

    def resize(self, w, h):
        self.setGeometry(0, 0, w, h)
        self.image_label.setGeometry(0, 0, w, h - 100)
        self.header_label.setGeometry(0, 290, w, h - 360)
        self.text_label.setGeometry(0, 340, w, h - 380)


class DropCountLabel(QtWidgets.QLabel):
    """
    While dragging files over the FileList, this counter displays the
    number of files you're dragging.
    """

    def __init__(self, common, parent):
        self.parent = parent
        super(DropCountLabel, self).__init__(parent=parent)

        self.common = common

        self.setAcceptDrops(True)
        self.setAlignment(QtCore.Qt.AlignCenter)
        self.setText(strings._("gui_drag_and_drop"))
        self.setStyleSheet(self.common.gui.css["share_file_selection_drop_count_label"])
        self.hide()

    def dragEnterEvent(self, event):
        self.hide()
        event.accept()


class FileList(QtWidgets.QListWidget):
    """
    The list of files and folders in the GUI.
    """

    files_dropped = QtCore.Signal()
    files_updated = QtCore.Signal()

    def __init__(self, common, background_image_filename, header_text, parent=None):
        super(FileList, self).__init__(parent)
        self.common = common
        self.setAcceptDrops(True)

        self.setIconSize(QtCore.QSize(32, 32))
        self.setSortingEnabled(True)
        self.setMinimumHeight(160)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.drop_here = DropHereWidget(
            self.common,
            background_image_filename,
            header_text,
            self.width(),
            self.height(),
            self,
        )
        self.drop_count = DropCountLabel(self.common, self)
        self.resizeEvent(None)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)

    def update(self):
        """
        Update the GUI elements based on the current state.
        """
        # file list should have a background image if empty
        if self.count() == 0:
            self.drop_here.show()
        else:
            self.drop_here.hide()

    def server_started(self):
        """
        Update the GUI when the server starts, by hiding delete buttons.
        """
        self.setAcceptDrops(False)
        self.setCurrentItem(None)
        self.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        for index in range(self.count()):
            self.item(index).item_button.hide()

    def server_stopped(self):
        """
        Update the GUI when the server stops, by showing delete buttons.
        """
        self.setAcceptDrops(True)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        for index in range(self.count()):
            self.item(index).item_button.show()

    def resizeEvent(self, event):
        """
        When the widget is resized, resize the drop files image and text.
        """
        self.drop_here.resize(self.width(), self.height())

        if self.count() > 0:
            # Add and delete an empty item, to force all items to get redrawn
            # This is ugly, but the only way I could figure out how to proceed
            item = QtWidgets.QListWidgetItem("fake item")
            self.addItem(item)
            self.takeItem(self.row(item))
            self.update()

            # Extend any filenames that were truncated to fit the window
            # We use 200 as a rough guess at how wide the 'file size + delete button' widget is
            # and extend based on the overall width minus that amount.
            for index in range(self.count()):
                metrics = QtGui.QFontMetrics(self.item(index).font())
                elided = metrics.elidedText(
                    self.item(index).basename, QtCore.Qt.ElideRight, self.width() - 200
                )
                self.item(index).setText(elided)

    def dragEnterEvent(self, event):
        """
        dragEnterEvent for dragging files and directories into the widget.
        """
        # Drag and drop doesn't work in Flatpak, because of the sandbox
        if self.common.is_flatpak():
            Alert(self.common, strings._("gui_dragdrop_sandbox_flatpak").format())
            event.ignore()
            return

        if event.mimeData().hasUrls:
            self.setStyleSheet(self.common.gui.css["share_file_list_drag_enter"])
            count = len(event.mimeData().urls())
            self.drop_count.setText(f"+{count}")

            size_hint = self.drop_count.sizeHint()
            self.drop_count.setGeometry(
                self.width() - size_hint.width() - 30,
                self.height() - size_hint.height() - 10,
                size_hint.width(),
                size_hint.height(),
            )
            self.drop_count.show()
            event.accept()
        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        """
        dragLeaveEvent for dragging files and directories into the widget.
        """
        # Drag and drop doesn't work in Flatpak, because of the sandbox
        if self.common.is_flatpak():
            event.ignore()
            return

        self.setStyleSheet(self.common.gui.css["share_file_list_drag_leave"])
        self.drop_count.hide()
        event.accept()
        self.update()

    def dragMoveEvent(self, event):
        """
        dragMoveEvent for dragging files and directories into the widget.
        """
        # Drag and drop doesn't work in Flatpak, because of the sandbox
        if self.common.is_flatpak():
            event.ignore()
            return

        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """
        dropEvent for dragging files and directories into the widget.
        """
        # Drag and drop doesn't work in Flatpak, because of the sandbox
        if self.common.is_flatpak():
            event.ignore()
            return

        if event.mimeData().hasUrls:
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
            for url in event.mimeData().urls():
                filename = str(url.toLocalFile())
                self.add_file(filename)
        else:
            event.ignore()

        self.setStyleSheet(self.common.gui.css["share_file_list_drag_leave"])
        self.drop_count.hide()

        self.files_dropped.emit()

    def add_file(self, filename):
        """
        Add a file or directory to this widget.
        """
        filenames = []
        for index in range(self.count()):
            filenames.append(self.item(index).filename)

        if filename not in filenames:
            if not os.access(filename, os.R_OK):
                Alert(self.common, strings._("not_a_readable_file").format(filename))
                return

            fileinfo = QtCore.QFileInfo(filename)
            ip = QtWidgets.QFileIconProvider()
            icon = ip.icon(fileinfo)

            if os.path.isfile(filename):
                size_bytes = fileinfo.size()
                size_readable = self.common.human_readable_filesize(size_bytes)
            else:
                size_bytes = self.common.dir_size(filename)
                size_readable = self.common.human_readable_filesize(size_bytes)

            # Create a new item
            item = QtWidgets.QListWidgetItem()
            item.setIcon(icon)
            item.size_bytes = size_bytes

            # Item's filename attribute and size labels
            item.filename = filename
            item_size = QtWidgets.QLabel(size_readable)
            item_size.setStyleSheet(self.common.gui.css["share_file_list_item_size"])

            item.basename = os.path.basename(filename.rstrip("/"))
            # Use the basename as the method with which to sort the list
            metrics = QtGui.QFontMetrics(item.font())
            elided = metrics.elidedText(
                item.basename, QtCore.Qt.ElideRight, self.sizeHint().width()
            )
            item.setData(QtCore.Qt.DisplayRole, elided)

            # Item's delete button
            def delete_item():
                itemrow = self.row(item)
                self.takeItem(itemrow)
                self.files_updated.emit()

            item.item_button = QtWidgets.QPushButton()
            item.item_button.setDefault(False)
            item.item_button.setFlat(True)
            item.item_button.setIcon(
                QtGui.QIcon(GuiCommon.get_resource_path("images/file_delete.png"))
            )
            item.item_button.clicked.connect(delete_item)
            item.item_button.setSizePolicy(
                QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed
            )

            # Item info widget, with a white background
            item_info_layout = QtWidgets.QHBoxLayout()
            item_info_layout.setContentsMargins(0, 0, 0, 0)
            item_info_layout.addWidget(item_size)
            item_info_layout.addWidget(item.item_button)
            item_info = QtWidgets.QWidget()
            item_info.setObjectName("item-info")
            item_info.setLayout(item_info_layout)

            # Create the item's widget and layouts
            item_hlayout = QtWidgets.QHBoxLayout()
            item_hlayout.addStretch()
            item_hlayout.addWidget(item_info)
            widget = QtWidgets.QWidget()
            widget.setLayout(item_hlayout)

            item.setSizeHint(widget.sizeHint())

            self.addItem(item)
            self.setItemWidget(item, widget)

            self.files_updated.emit()


class FileSelection(QtWidgets.QVBoxLayout):
    """
    The list of files and folders in the GUI, as well as buttons to add and
    delete the files and folders.
    """

    def __init__(self, common, background_image_filename, header_text, parent):
        super(FileSelection, self).__init__()

        self.common = common
        self.parent = parent

        self.server_on = False

        # File list
        self.file_list = FileList(self.common, background_image_filename, header_text)
        self.file_list.itemSelectionChanged.connect(self.update)
        self.file_list.files_dropped.connect(self.update)
        self.file_list.files_updated.connect(self.update)

        # Sandboxes (for masOS, Flatpak, etc.) need separate add files and folders buttons, in
        # order to use native file selection dialogs
        if self.common.platform == "Darwin" or self.common.is_flatpak():
            self.sandbox = True
        else:
            self.sandbox = False

        # Buttons
        if self.sandbox:
            # The macOS sandbox makes it so the Mac version needs separate add files
            # and folders buttons, in order to use native file selection dialogs
            self.add_files_button = QtWidgets.QPushButton(strings._("gui_add_files"))
            self.add_files_button.clicked.connect(self.add_files)
            self.add_folder_button = QtWidgets.QPushButton(strings._("gui_add_folder"))
            self.add_folder_button.clicked.connect(self.add_folder)
        else:
            self.add_button = QtWidgets.QPushButton(strings._("gui_add"))
            self.add_button.clicked.connect(self.add)
        self.remove_button = QtWidgets.QPushButton(strings._("gui_remove"))
        self.remove_button.clicked.connect(self.delete)
        button_layout = QtWidgets.QHBoxLayout()
        button_layout.addStretch()
        if self.sandbox:
            button_layout.addWidget(self.add_files_button)
            button_layout.addWidget(self.add_folder_button)
        else:
            button_layout.addWidget(self.add_button)
        button_layout.addWidget(self.remove_button)

        # Add the widgets
        self.addWidget(self.file_list)
        self.addLayout(button_layout)

        self.update()

    def update(self):
        """
        Update the GUI elements based on the current state.
        """
        # All buttons should be hidden if the server is on
        if self.server_on:
            if self.sandbox:
                self.add_files_button.hide()
                self.add_folder_button.hide()
            else:
                self.add_button.hide()
            self.remove_button.hide()
        else:
            if self.sandbox:
                self.add_files_button.show()
                self.add_folder_button.show()
            else:
                self.add_button.show()

            # Delete button should be hidden if item isn't selected
            if len(self.file_list.selectedItems()) == 0:
                self.remove_button.hide()
            else:
                self.remove_button.show()

        # Update the file list
        self.file_list.update()

        # Save the latest file list to mode settings
        self.save_filenames()

    def add(self):
        """
        Add button clicked.
        """
        file_dialog = AddFileDialog(self.common, caption=strings._("gui_choose_items"))
        if file_dialog.exec() == QtWidgets.QDialog.Accepted:
            self.common.log("FileSelection", "add", file_dialog.selectedFiles())
            for filename in file_dialog.selectedFiles():
                self.file_list.add_file(filename)

        self.file_list.setCurrentItem(None)
        self.update()

    def add_files(self):
        """
        Add Files button clicked.
        """
        files = QtWidgets.QFileDialog.getOpenFileNames(
            self.parent, caption=strings._("gui_choose_items")
        )
        self.common.log("FileSelection", "add_files", files)

        filenames = files[0]
        for filename in filenames:
            self.file_list.add_file(filename)

        self.file_list.setCurrentItem(None)
        self.update()

    def add_folder(self):
        """
        Add Folder button clicked.
        """
        filename = QtWidgets.QFileDialog.getExistingDirectory(
            self.parent,
            caption=strings._("gui_choose_items"),
            options=QtWidgets.QFileDialog.ShowDirsOnly,
        )
        self.common.log("FileSelection", "add_folder", filename)
        if filename:
            self.file_list.add_file(filename)
            self.file_list.setCurrentItem(None)
            self.update()

    def delete(self):
        """
        Delete button clicked
        """
        selected = self.file_list.selectedItems()
        for item in selected:
            itemrow = self.file_list.row(item)
            self.file_list.takeItem(itemrow)
        self.file_list.files_updated.emit()

        self.file_list.setCurrentItem(None)
        self.update()

    def server_started(self):
        """
        Gets called when the server starts.
        """
        self.server_on = True
        self.file_list.server_started()
        self.update()

    def server_stopped(self):
        """
        Gets called when the server stops.
        """
        self.server_on = False
        self.file_list.server_stopped()
        self.update()

    def get_num_files(self):
        """
        Returns the total number of files and folders in the list.
        """
        return len(range(self.file_list.count()))

    def get_filenames(self):
        """
        Return the list of file and folder names
        """
        filenames = []
        for index in range(self.file_list.count()):
            filenames.append(self.file_list.item(index).filename)
        return filenames

    def save_filenames(self):
        """
        Save the filenames to mode settings
        """
        filenames = self.get_filenames()
        if self.parent.tab.mode == self.common.gui.MODE_SHARE:
            self.parent.settings.set("share", "filenames", filenames)
        elif self.parent.tab.mode == self.common.gui.MODE_WEBSITE:
            self.parent.settings.set("website", "filenames", filenames)

    def setFocus(self):
        """
        Set the Qt app focus on the file selection box.
        """
        self.file_list.setFocus()



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from PySide6 import QtCore, QtWidgets, QtGui
import requests
import os
import base64
import json

from . import strings
from .gui_common import GuiCommon
from onionshare_cli.meek import MeekNotFound, MeekNotRunning


class MoatDialog(QtWidgets.QDialog):
    """
    Moat dialog: Request a bridge from torproject.org
    """

    got_bridges = QtCore.Signal(str)

    def __init__(self, common, meek):
        super(MoatDialog, self).__init__()

        self.common = common

        self.common.log("MoatDialog", "__init__")

        self.meek = meek

        self.setModal(True)
        self.setWindowTitle(strings._("gui_settings_bridge_moat_button"))
        self.setWindowIcon(QtGui.QIcon(GuiCommon.get_resource_path("images/logo.png")))

        # Label
        self.label = QtWidgets.QLabel()

        # CAPTCHA image
        self.captcha = QtWidgets.QLabel()
        self.captcha.setFixedSize(400, 125)  # this is the size of the CAPTCHA image

        # Solution input
        self.solution_lineedit = QtWidgets.QLineEdit()
        self.solution_lineedit.setPlaceholderText(strings._("moat_captcha_placeholder"))
        self.solution_lineedit.editingFinished.connect(
            self.solution_lineedit_editing_finished
        )
        self.submit_button = QtWidgets.QPushButton(strings._("moat_captcha_submit"))
        self.submit_button.clicked.connect(self.submit_clicked)
        solution_layout = QtWidgets.QHBoxLayout()
        solution_layout.addWidget(self.solution_lineedit)
        solution_layout.addWidget(self.submit_button)

        # Error label
        self.error_label = QtWidgets.QLabel()
        self.error_label.setStyleSheet(self.common.gui.css["tor_settings_error"])
        self.error_label.hide()

        # Buttons
        self.reload_button = QtWidgets.QPushButton(strings._("moat_captcha_reload"))
        self.reload_button.clicked.connect(self.reload_clicked)
        self.cancel_button = QtWidgets.QPushButton(
            strings._("gui_settings_button_cancel")
        )
        self.cancel_button.clicked.connect(self.cancel_clicked)
        buttons_layout = QtWidgets.QHBoxLayout()
        buttons_layout.addStretch()
        buttons_layout.addWidget(self.reload_button)
        buttons_layout.addWidget(self.cancel_button)

        # Layout
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.label)
        layout.addWidget(self.captcha)
        layout.addLayout(solution_layout)
        layout.addStretch()
        layout.addWidget(self.error_label)
        layout.addLayout(buttons_layout)

        self.setLayout(layout)
        self.cancel_button.setFocus()

        self.reload_clicked()

    def reload_clicked(self):
        """
        Reload button clicked.
        """
        self.common.log("MoatDialog", "reload_clicked")

        self.label.setText(strings._("moat_contact_label"))
        self.error_label.hide()

        self.captcha.hide()
        self.solution_lineedit.hide()
        self.reload_button.hide()
        self.submit_button.hide()

        # BridgeDB fetch
        self.t_fetch = MoatThread(self.common, self.meek, "fetch")
        self.t_fetch.bridgedb_error.connect(self.bridgedb_error)
        self.t_fetch.captcha_ready.connect(self.captcha_ready)
        self.t_fetch.start()

    def submit_clicked(self):
        """
        Submit button clicked.
        """
        self.error_label.hide()
        self.solution_lineedit.setEnabled(False)

        solution = self.solution_lineedit.text().strip()
        if len(solution) == 0:
            self.common.log("MoatDialog", "submit_clicked", "solution is blank")
            self.error_label.setText(strings._("moat_solution_empty_error"))
            self.error_label.show()
            return

        # BridgeDB check
        self.t_check = MoatThread(
            self.common,
            self.meek,
            "check",
            {
                "transport": self.transport,
                "challenge": self.challenge,
                "solution": self.solution_lineedit.text(),
            },
        )
        self.t_check.bridgedb_error.connect(self.bridgedb_error)
        self.t_check.captcha_error.connect(self.captcha_error)
        self.t_check.bridges_ready.connect(self.bridges_ready)
        self.t_check.start()

    def cancel_clicked(self):
        """
        Cancel button clicked.
        """
        self.common.log("MoatDialog", "cancel_clicked")
        self.close()

    def bridgedb_error(self):
        self.common.log("MoatDialog", "bridgedb_error")
        self.error_label.setText(strings._("moat_bridgedb_error"))
        self.error_label.show()

        self.solution_lineedit.setEnabled(True)

    def captcha_error(self, msg):
        self.common.log("MoatDialog", "captcha_error")
        if msg == "":
            self.error_label.setText(strings._("moat_captcha_error"))
        else:
            self.error_label.setText(msg)
        self.error_label.show()

        self.solution_lineedit.setEnabled(True)

    def captcha_ready(self, transport, image, challenge):
        self.common.log("MoatDialog", "captcha_ready")

        self.transport = transport
        self.challenge = challenge

        # Save captcha image to disk, so we can load it
        captcha_data = base64.b64decode(image)
        captcha_filename = os.path.join(self.common.build_tmp_dir(), "captcha.jpg")
        with open(captcha_filename, "wb") as f:
            f.write(captcha_data)

        self.captcha.setPixmap(QtGui.QPixmap.fromImage(QtGui.QImage(captcha_filename)))
        os.remove(captcha_filename)

        self.label.setText(strings._("moat_captcha_label"))
        self.captcha.show()
        self.solution_lineedit.setEnabled(True)
        self.solution_lineedit.setText("")
        self.solution_lineedit.show()
        self.solution_lineedit.setFocus()
        self.reload_button.show()
        self.submit_button.show()

    def solution_lineedit_editing_finished(self):
        self.common.log("MoatDialog", "solution_lineedit_editing_finished")

    def bridges_ready(self, bridges):
        self.common.log("MoatDialog", "bridges_ready", bridges)
        self.got_bridges.emit(bridges)
        self.close()


class MoatThread(QtCore.QThread):
    """
    This does all of the communicating with BridgeDB in a separate thread.

    Valid actions are:
    - "fetch": requests a new CAPTCHA
    - "check": sends a CAPTCHA solution

    """

    bridgedb_error = QtCore.Signal()
    captcha_error = QtCore.Signal(str)
    captcha_ready = QtCore.Signal(str, str, str)
    bridges_ready = QtCore.Signal(str)

    def __init__(self, common, meek, action, data={}):
        super(MoatThread, self).__init__()
        self.common = common
        self.common.log("MoatThread", "__init__", f"action={action}")

        self.meek = meek
        self.transport = "obfs4"
        self.action = action
        self.data = data

    def run(self):

        # Start Meek so that we can do domain fronting
        try:
            self.meek.start()
        except (
            MeekNotFound,
            MeekNotRunning,
        ):
            self.bridgedb_error.emit()
            return

        # We should only fetch bridges if we can domain front,
        # but we can override this in local-only mode.
        if not self.meek.meek_proxies and not self.common.gui.local_only:
            self.common.log(
                "MoatThread", "run", f"Could not identify meek proxies to make request"
            )
            self.bridgedb_error.emit()
            return

        if self.action == "fetch":
            self.common.log("MoatThread", "run", f"starting fetch")

            # Request a bridge
            r = requests.post(
                "https://bridges.torproject.org/moat/fetch",
                headers={"Content-Type": "application/vnd.api+json"},
                proxies=self.meek.meek_proxies,
                json={
                    "data": [
                        {
                            "version": "0.1.0",
                            "type": "client-transports",
                            "supported": ["obfs4", "snowflake"],
                        }
                    ]
                },
            )

            self.meek.cleanup()

            if r.status_code != 200:
                self.common.log("MoatThread", "run", f"status_code={r.status_code}")
                self.bridgedb_error.emit()
                return

            try:
                moat_res = r.json()
                if "errors" in moat_res:
                    self.common.log("MoatThread", "run", f"errors={moat_res['errors']}")
                    self.bridgedb_error.emit()
                    return
                if "data" not in moat_res:
                    self.common.log("MoatThread", "run", f"no data")
                    self.bridgedb_error.emit()
                    return
                if moat_res["data"][0]["type"] != "moat-challenge":
                    self.common.log("MoatThread", "run", f"type != moat-challenge")
                    self.bridgedb_error.emit()
                    return

                transport = moat_res["data"][0]["transport"]
                image = moat_res["data"][0]["image"]
                challenge = moat_res["data"][0]["challenge"]

                self.captcha_ready.emit(transport, image, challenge)
            except Exception as e:
                self.common.log("MoatThread", "run", f"hit exception: {e}")
                self.bridgedb_error.emit()
                return

        elif self.action == "check":
            self.common.log("MoatThread", "run", f"starting check")

            # Check the CAPTCHA
            r = requests.post(
                "https://bridges.torproject.org/moat/check",
                headers={"Content-Type": "application/vnd.api+json"},
                proxies=self.meek.meek_proxies,
                json={
                    "data": [
                        {
                            "id": "2",
                            "type": "moat-solution",
                            "version": "0.1.0",
                            "transport": self.data["transport"],
                            "challenge": self.data["challenge"],
                            "solution": self.data["solution"],
                            "qrcode": "false",
                        }
                    ]
                },
            )

            self.meek.cleanup()

            if r.status_code != 200:
                self.common.log("MoatThread", "run", f"status_code={r.status_code}")
                self.bridgedb_error.emit()
                return

            try:
                moat_res = r.json()
                self.common.log(
                    "MoatThread",
                    "run",
                    f"got bridges:\n{json.dumps(moat_res,indent=2)}",
                )

                if "errors" in moat_res:
                    self.common.log("MoatThread", "run", f"errors={moat_res['errors']}")
                    if moat_res["errors"][0]["code"] == 419:
                        self.captcha_error.emit("")
                        return
                    else:
                        errors = " ".join([e["detail"] for e in moat_res["errors"]])
                        self.captcha_error.emit(errors)
                        return

                if moat_res["data"][0]["type"] != "moat-bridges":
                    self.common.log("MoatThread", "run", f"type != moat-bridges")
                    self.bridgedb_error.emit()
                    return

                bridges = moat_res["data"][0]["bridges"]
                self.bridges_ready.emit("\n".join(bridges))

            except Exception as e:
                self.common.log("MoatThread", "run", f"hit exception: {e}")
                self.bridgedb_error.emit()
                return

        else:
            self.common.log("MoatThread", "run", f"invalid action: {self.action}")



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
import os
import sys
import argparse
import signal
import json
import psutil
import getpass
from PySide6 import QtCore, QtWidgets, QtGui

from PySide6.QtCore import Slot, Qt
from PySide6.QtGui import QPalette, QColor

from onionshare_cli.common import Common
from onionshare_cli.settings import Settings

from .gui_common import GuiCommon
from .widgets import Alert
from .main_window import MainWindow


class Application(QtWidgets.QApplication):
    """
    This is Qt's QApplication class. It has been overridden to support threads
    and the quick keyboard shortcut.
    """

    def __init__(self, common):
        if common.platform == "Linux" or common.platform == "BSD":
            self.setDesktopFileName("org.onionshare.OnionShare")
            self.setOrganizationDomain("org.onionshare.OnionShare")
            self.setOrganizationName("OnionShare")

        QtWidgets.QApplication.__init__(self, sys.argv)

        if common.platform == "Darwin" or common.platform == "Windows":
            self.setStyle("Fusion")

        # Check color mode on starting the app
        self.color_mode = self.get_color_mode(common)

        # Enable Dark Theme
        if self.color_mode == "dark":
            self.setDarkMode()
        else:
            self.setLightMode()

        self.installEventFilter(self)

    def eventFilter(self, obj, event):
        if (
            event.type() == QtCore.QEvent.KeyPress
            and event.key() == QtCore.Qt.Key_Q
            and event.modifiers() == QtCore.Qt.ControlModifier
        ):
            self.quit()
        return False

    def is_dark_mode(self):
        baseColor = QtGui.QPalette().color(QtGui.QPalette.Base)
        if baseColor.name().lower() == "#ffffff":
            return False
        return True

    def setLightMode(self):
        light_palette = QPalette()
        light_palette.setColor(QPalette.Window, QColor(236, 236, 236))
        light_palette.setColor(QPalette.WindowText, Qt.black)
        light_palette.setColor(QPalette.Base, Qt.white)
        light_palette.setColor(QPalette.AlternateBase, QColor(245, 245, 245))
        light_palette.setColor(QPalette.ToolTipBase, Qt.white)
        light_palette.setColor(QPalette.ToolTipText, Qt.black)
        light_palette.setColor(QPalette.Text, Qt.black)
        light_palette.setColor(QPalette.Button, QColor(236, 236, 236))
        light_palette.setColor(QPalette.ButtonText, Qt.black)
        light_palette.setColor(QPalette.BrightText, Qt.white)
        light_palette.setColor(QPalette.Link, QColor(0, 104, 218))
        light_palette.setColor(QPalette.Highlight, QColor(179, 215, 255))
        light_palette.setColor(QPalette.HighlightedText, Qt.black)
        light_palette.setColor(QPalette.Light, Qt.white)
        light_palette.setColor(QPalette.Midlight, QColor(245, 245, 245))
        light_palette.setColor(QPalette.Dark, QColor(191, 191, 191))
        light_palette.setColor(QPalette.Mid, QColor(169, 169, 169))
        light_palette.setColor(QPalette.Shadow, Qt.black)
        self.setPalette(light_palette)

    def setDarkMode(self):
        dark_palette = QPalette()
        dark_palette.setColor(QPalette.Window, QColor(53, 53, 53))
        dark_palette.setColor(QPalette.WindowText, Qt.white)
        dark_palette.setColor(QPalette.Base, QColor(25, 25, 25))
        dark_palette.setColor(QPalette.AlternateBase, QColor(53, 53, 53))
        dark_palette.setColor(QPalette.ToolTipBase, Qt.white)
        dark_palette.setColor(QPalette.ToolTipText, Qt.white)
        dark_palette.setColor(QPalette.Text, Qt.white)
        dark_palette.setColor(QPalette.Button, QColor(53, 53, 53))
        dark_palette.setColor(QPalette.ButtonText, Qt.white)
        dark_palette.setColor(QPalette.BrightText, Qt.red)
        dark_palette.setColor(QPalette.Link, QColor(42, 130, 218))
        dark_palette.setColor(QPalette.Highlight, QColor(42, 130, 218))
        dark_palette.setColor(QPalette.HighlightedText, Qt.black)
        self.setPalette(dark_palette)
        self.setStyleSheet(
            "QToolTip { color: #ffffff; background-color: #2a82da; border: 1px solid white; }"
        )

    def get_color_mode(self, common):
        curr_settings = Settings(common)
        curr_settings.load()
        current_theme = curr_settings.get("theme")

        if current_theme == 1:
            return "light"
        elif current_theme == 2:
            return "dark"
        else:
            return "dark" if self.is_dark_mode() else "light"


def main():
    """
    The main() function implements all of the logic that the GUI version of onionshare uses.
    """
    common = Common()
    common.display_banner()

    # Start the Qt app
    global qtapp
    qtapp = Application(common)

    # Parse arguments
    parser = argparse.ArgumentParser(
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=48)
    )
    parser.add_argument(
        "--local-only",
        action="store_true",
        dest="local_only",
        help="Don't use Tor (only for development)",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        help="Log OnionShare errors to stdout, and web errors to disk",
    )
    parser.add_argument(
        "--filenames",
        metavar="filenames",
        nargs="+",
        help="List of files or folders to share",
    )
    args = parser.parse_args()

    filenames = args.filenames
    if filenames:
        for i in range(len(filenames)):
            filenames[i] = os.path.abspath(filenames[i])

    local_only = bool(args.local_only)
    verbose = bool(args.verbose)

    # Verbose mode?
    common.verbose = verbose

    # Attach the GUI common parts to the common object
    common.gui = GuiCommon(common, qtapp, local_only)

    # Validation
    if filenames:
        valid = True
        for filename in filenames:
            if not os.path.isfile(filename) and not os.path.isdir(filename):
                Alert(common, f"{filename} is not a valid file.")
                valid = False
            if not os.access(filename, os.R_OK):
                Alert(common, f"{filename} is not a readable file.")
                valid = False
        if not valid:
            sys.exit()

    # Is there another onionshare-gui running?
    if os.path.exists(common.gui.lock_filename):
        with open(common.gui.lock_filename, "r") as f:
            existing_pid = int(f.read())

        # Is this process actually still running?
        still_running = True
        if not psutil.pid_exists(existing_pid):
            still_running = False
        else:
            for proc in psutil.process_iter(["pid", "name", "username"]):
                if proc.pid == existing_pid:
                    if (
                        proc.username() != getpass.getuser()
                        or "onionshare" not in " ".join(proc.cmdline()).lower()
                    ):
                        still_running = False

        if still_running:
            print(f"Opening tab in existing OnionShare window (pid {existing_pid})")

            # Make an event for the existing OnionShare window
            if filenames:
                obj = {"type": "new_share_tab", "filenames": filenames}
            else:
                obj = {"type": "new_tab"}

            # Write that event to disk
            with open(common.gui.events_filename, "a") as f:
                f.write(json.dumps(obj) + "\n")
            return
        else:
            os.remove(common.gui.lock_filename)

    # Write the lock file
    with open(common.gui.lock_filename, "w") as f:
        f.write(f"{os.getpid()}\n")

    # Allow Ctrl-C to smoothly quit the program instead of throwing an exception
    def signal_handler(s, frame):
        print("\nCtrl-C pressed, quitting")
        if os.path.exists(common.gui.lock_filename):
            os.remove(common.gui.lock_filename)
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)

    # Launch the gui
    main_window = MainWindow(common, filenames)

    # If filenames were passed in, open them in a tab
    if filenames:
        main_window.tabs.new_share_tab(filenames)

    # Clean up when app quits
    def shutdown():
        main_window.cleanup()
        os.remove(common.gui.lock_filename)

    qtapp.aboutToQuit.connect(shutdown)

    # All done
    sys.exit(qtapp.exec())



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
from PySide6 import QtCore, QtWidgets, QtGui

from onionshare_cli.web import Web

from ..history import History, ToggleHistory, ReceiveHistoryItem
from .. import Mode
from .... import strings
from ....widgets import MinimumSizeWidget, Alert
from ....gui_common import GuiCommon


class ReceiveMode(Mode):
    """
    Parts of the main window UI for receiving files.
    """

    def init(self):
        """
        Custom initialization for ReceiveMode.
        """
        # Create the Web object
        self.web = Web(self.common, True, self.settings, "receive")

        # Receive image
        self.image_label = QtWidgets.QLabel()
        self.image_label.setPixmap(
            QtGui.QPixmap.fromImage(
                QtGui.QImage(
                    GuiCommon.get_resource_path(
                        "images/{}_mode_receive.png".format(self.common.gui.color_mode)
                    )
                )
            )
        )
        self.image_label.setFixedSize(250, 250)
        image_layout = QtWidgets.QVBoxLayout()
        image_layout.addWidget(self.image_label)
        self.image = QtWidgets.QWidget()
        self.image.setLayout(image_layout)

        # Settings

        # Data dir
        data_dir_label = QtWidgets.QLabel(
            strings._("mode_settings_receive_data_dir_label")
        )
        self.data_dir_lineedit = QtWidgets.QLineEdit()
        self.data_dir_lineedit.setReadOnly(True)
        self.data_dir_lineedit.setText(self.settings.get("receive", "data_dir"))
        data_dir_button = QtWidgets.QPushButton(
            strings._("mode_settings_receive_data_dir_browse_button")
        )
        data_dir_button.clicked.connect(self.data_dir_button_clicked)
        data_dir_layout = QtWidgets.QHBoxLayout()
        data_dir_layout.addWidget(data_dir_label)
        data_dir_layout.addWidget(self.data_dir_lineedit)
        data_dir_layout.addWidget(data_dir_button)
        self.mode_settings_widget.mode_specific_layout.addLayout(data_dir_layout)

        # Disable text or files
        self.disable_text_checkbox = self.settings.get("receive", "disable_files")
        self.disable_text_checkbox = QtWidgets.QCheckBox()
        self.disable_text_checkbox.clicked.connect(self.disable_text_checkbox_clicked)
        self.disable_text_checkbox.setText(
            strings._("mode_settings_receive_disable_text_checkbox")
        )
        self.disable_text_checkbox.setStyleSheet(self.common.gui.css["receive_options"])
        self.disable_files_checkbox = self.settings.get("receive", "disable_files")
        self.disable_files_checkbox = QtWidgets.QCheckBox()
        self.disable_files_checkbox.clicked.connect(self.disable_files_checkbox_clicked)
        self.disable_files_checkbox.setText(
            strings._("mode_settings_receive_disable_files_checkbox")
        )
        self.disable_files_checkbox.setStyleSheet(self.common.gui.css["receive_options"])
        disable_layout = QtWidgets.QHBoxLayout()
        disable_layout.addWidget(self.disable_text_checkbox)
        disable_layout.addWidget(self.disable_files_checkbox)
        disable_layout.addStretch()
        self.mode_settings_widget.mode_specific_layout.addLayout(disable_layout)

        # Webhook URL
        webhook_url = self.settings.get("receive", "webhook_url")
        self.webhook_url_checkbox = QtWidgets.QCheckBox()
        self.webhook_url_checkbox.clicked.connect(self.webhook_url_checkbox_clicked)
        self.webhook_url_checkbox.setText(
            strings._("mode_settings_receive_webhook_url_checkbox")
        )
        self.webhook_url_lineedit = QtWidgets.QLineEdit()
        self.webhook_url_lineedit.editingFinished.connect(
            self.webhook_url_editing_finished
        )
        self.webhook_url_lineedit.setPlaceholderText(
            "https://example.com/post-when-file-uploaded"
        )
        webhook_url_layout = QtWidgets.QHBoxLayout()
        webhook_url_layout.addWidget(self.webhook_url_checkbox)
        webhook_url_layout.addWidget(self.webhook_url_lineedit)
        if webhook_url is not None and webhook_url != "":
            self.webhook_url_checkbox.setCheckState(QtCore.Qt.Checked)
            self.webhook_url_lineedit.setText(
                self.settings.get("receive", "webhook_url")
            )
            self.show_webhook_url()
        else:
            self.webhook_url_checkbox.setCheckState(QtCore.Qt.Unchecked)
            self.hide_webhook_url()
        self.mode_settings_widget.mode_specific_layout.addLayout(webhook_url_layout)

        # Set title placeholder
        self.mode_settings_widget.title_lineedit.setPlaceholderText(
            strings._("gui_tab_name_receive")
        )

        # Server status
        self.server_status.set_mode("receive")
        self.server_status.server_started_finished.connect(self.update_primary_action)
        self.server_status.server_stopped.connect(self.update_primary_action)
        self.server_status.server_canceled.connect(self.update_primary_action)

        # Tell server_status about web, then update
        self.server_status.web = self.web
        self.server_status.update()

        # Upload history
        self.history = History(
            self.common,
            QtGui.QPixmap.fromImage(
                QtGui.QImage(
                    GuiCommon.get_resource_path("images/receive_icon_transparent.png")
                )
            ),
            strings._("gui_receive_mode_no_files"),
            strings._("gui_all_modes_history"),
        )
        self.history.hide()

        # Toggle history
        self.toggle_history = ToggleHistory(
            self.common,
            self,
            self.history,
            QtGui.QIcon(GuiCommon.get_resource_path(f"images/{self.common.gui.color_mode}_history_icon_toggle.svg")),
            QtGui.QIcon(
                GuiCommon.get_resource_path(f"images/{self.common.gui.color_mode}_history_icon_toggle_selected.svg")
            ),
        )

        # Header
        header_label = QtWidgets.QLabel(strings._("gui_new_tab_receive_button"))
        header_label.setStyleSheet(self.common.gui.css["mode_header_label"])

        # Receive mode warning
        receive_warning = QtWidgets.QLabel(strings._("gui_receive_mode_warning"))
        receive_warning.setMinimumHeight(80)
        receive_warning.setWordWrap(True)

        # Top bar
        top_bar_layout = QtWidgets.QHBoxLayout()
        top_bar_layout.addStretch()
        top_bar_layout.addWidget(self.toggle_history)

        # Main layout
        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.addWidget(header_label)
        self.main_layout.addWidget(receive_warning)
        self.main_layout.addWidget(self.primary_action, stretch=1)
        self.main_layout.addWidget(self.server_status)

        # Row layout
        content_row = QtWidgets.QHBoxLayout()
        content_row.addLayout(self.main_layout, stretch=1)
        content_row.addWidget(self.image)
        row_layout = QtWidgets.QVBoxLayout()
        row_layout.addLayout(top_bar_layout)
        row_layout.addLayout(content_row, stretch=1)

        # Column layout
        self.column_layout = QtWidgets.QHBoxLayout()
        self.column_layout.addLayout(row_layout)
        self.column_layout.addWidget(self.history, stretch=1)

        # Content layout
        self.content_layout.addLayout(self.column_layout)

    def get_type(self):
        """
        Returns the type of mode as a string (e.g. "share", "receive", etc.)
        """
        return "receive"

    def data_dir_button_clicked(self):
        """
        Browse for a new OnionShare data directory, and save to tab settings
        """
        data_dir = self.data_dir_lineedit.text()
        selected_dir = QtWidgets.QFileDialog.getExistingDirectory(
            self, strings._("mode_settings_receive_data_dir_label"), data_dir
        )

        if selected_dir:
            # If we're running inside a flatpak package, the data dir must be inside ~/OnionShare
            if self.common.gui.is_flatpak:
                if not selected_dir.startswith(os.path.expanduser("~/OnionShare")):
                    Alert(self.common, strings._("gui_receive_flatpak_data_dir"))
                    return

            self.common.log(
                "ReceiveMode",
                "data_dir_button_clicked",
                f"selected dir: {selected_dir}",
            )
            self.data_dir_lineedit.setText(selected_dir)
            self.settings.set("receive", "data_dir", selected_dir)

    def disable_text_checkbox_clicked(self):
        self.settings.set(
            "receive", "disable_text", self.disable_text_checkbox.isChecked()
        )
        if self.disable_text_checkbox.isChecked():
            # Prevent also disabling files if text is disabled
            self.disable_files_checkbox.setDisabled(True)
        else:
            self.disable_files_checkbox.setDisabled(False)

    def disable_files_checkbox_clicked(self):
        self.settings.set(
            "receive", "disable_files", self.disable_files_checkbox.isChecked()
        )
        if self.disable_files_checkbox.isChecked():
            # Prevent also disabling text if files is disabled
            self.disable_text_checkbox.setDisabled(True)
        else:
            self.disable_text_checkbox.setDisabled(False)

    def webhook_url_checkbox_clicked(self):
        if self.webhook_url_checkbox.isChecked():
            if self.settings.get("receive", "webhook_url"):
                self.webhook_url_lineedit.setText(
                    self.settings.get("receive", "webhook_url")
                )
            self.show_webhook_url()
        else:
            self.settings.set("receive", "webhook_url", None)
            self.hide_webhook_url()

    def webhook_url_editing_finished(self):
        self.settings.set("receive", "webhook_url", self.webhook_url_lineedit.text())

    def hide_webhook_url(self):
        self.webhook_url_lineedit.hide()

    def show_webhook_url(self):
        self.webhook_url_lineedit.show()

    def get_stop_server_autostop_timer_text(self):
        """
        Return the string to put on the stop server button, if there's an auto-stop timer
        """
        return strings._("gui_receive_stop_server_autostop_timer")

    def autostop_timer_finished_should_stop_server(self):
        """
        The auto-stop timer expired, should we stop the server? Returns a bool
        """
        # If there were no attempts to upload files, or all uploads are done, we can stop
        if (
            self.web.receive_mode.cur_history_id == 0
            or not self.web.receive_mode.uploads_in_progress
        ):
            self.server_status.stop_server()
            self.server_status_label.setText(strings._("close_on_autostop_timer"))
            return True
        # An upload is probably still running - hold off on stopping the share, but block new shares.
        else:
            self.server_status_label.setText(
                strings._("gui_receive_mode_autostop_timer_waiting")
            )
            self.web.receive_mode.can_upload = False
            return False

    def start_server_custom(self):
        """
        Starting the server.
        """
        # Reset web counters
        self.web.receive_mode.cur_history_id = 0

        # Hide and reset the uploads if we have previously shared
        self.reset_info_counters()

        # Set proxies for webhook URL
        if self.common.gui.local_only:
            self.web.proxies = None
        else:
            (socks_address, socks_port) = self.common.gui.onion.get_tor_socks_port()
            self.web.proxies = {
                "http": f"socks5h://{socks_address}:{socks_port}",
                "https": f"socks5h://{socks_address}:{socks_port}",
            }

    def start_server_step2_custom(self):
        """
        Step 2 in starting the server.
        """
        # Continue
        self.starting_server_step3.emit()
        self.start_server_finished.emit()

    def handle_tor_broke_custom(self):
        """
        Connection to Tor broke.
        """
        self.primary_action.hide()

    def handle_request_load(self, event):
        """
        Handle REQUEST_LOAD event.
        """
        self.system_tray.showMessage(
            strings._("systray_page_loaded_title"),
            strings._("systray_page_loaded_message"),
        )

    def handle_request_started(self, event):
        """
        Handle REQUEST_STARTED event.
        """
        item = ReceiveHistoryItem(
            self.common,
            event["data"]["id"],
            event["data"]["content_length"],
        )

        self.history.add(event["data"]["id"], item)
        self.toggle_history.update_indicator(True)
        self.history.in_progress_count += 1
        self.history.update_in_progress()

        self.system_tray.showMessage(
            strings._("systray_receive_started_title"),
            strings._("systray_receive_started_message"),
        )

    def handle_request_progress(self, event):
        """
        Handle REQUEST_PROGRESS event.
        """
        self.history.update(
            event["data"]["id"],
            {"action": "progress", "progress": event["data"]["progress"]},
        )

    def handle_request_upload_includes_message(self, event):
        """
        Handle REQUEST_UPLOAD_INCLUDES_MESSAGE event.
        """
        self.history.includes_message(event["data"]["id"], event["data"]["filename"])

    def handle_request_upload_file_renamed(self, event):
        """
        Handle REQUEST_UPLOAD_FILE_RENAMED event.
        """
        self.history.update(
            event["data"]["id"],
            {
                "action": "rename",
                "old_filename": event["data"]["old_filename"],
                "new_filename": event["data"]["new_filename"],
            },
        )

    def handle_request_upload_set_dir(self, event):
        """
        Handle REQUEST_UPLOAD_SET_DIR event.
        """
        self.history.update(
            event["data"]["id"],
            {
                "action": "set_dir",
                "filename": event["data"]["filename"],
                "dir": event["data"]["dir"],
            },
        )

    def handle_request_upload_finished(self, event):
        """
        Handle REQUEST_UPLOAD_FINISHED event.
        """
        self.history.update(event["data"]["id"], {"action": "finished"})
        self.history.completed_count += 1
        self.history.in_progress_count -= 1
        self.history.update_completed()
        self.history.update_in_progress()

    def handle_request_upload_canceled(self, event):
        """
        Handle REQUEST_UPLOAD_CANCELED event.
        """
        self.history.update(event["data"]["id"], {"action": "canceled"})
        self.history.in_progress_count -= 1
        self.history.update_in_progress()

    def reset_info_counters(self):
        """
        Set the info counters back to zero.
        """
        self.history.reset()
        self.toggle_history.indicator_count = 0
        self.toggle_history.update_indicator()

    def update_primary_action(self):
        self.common.log("ReceiveMode", "update_primary_action")



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import json
import os
import locale


class Settings(object):
    """
    This class stores all of the settings for OnionShare, specifically for how
    to connect to Tor. If it can't find the settings file, it uses the default,
    which is to attempt to connect automatically using default Tor Browser
    settings.
    """

    def __init__(self, common, config=False):
        self.common = common

        self.common.log("Settings", "__init__")

        # If a readable config file was provided, use that instead
        if config:
            if os.path.isfile(config):
                self.filename = config
            else:
                self.common.log(
                    "Settings",
                    "__init__",
                    "Supplied config does not exist or is unreadable. Falling back to default location",
                )
                self.filename = self.build_filename()

        else:
            # Default config
            self.filename = self.build_filename()

        # Dictionary of available languages in this version of OnionShare,
        # mapped to the language name, in that language
        self.available_locales = {
            "af": "Afrikaans",  # Afrikaans
            "sq": "Shqip",  # Albanian
            "ar": "العربية",  # Arabic
            "be": "Беларуская",  # Belarusian
            "bn": "বাংলা",  # Bengali
            "bg": "Български",  #Bulgarian
            "ca": "Català",  # Catalan
            "zh_Hant": "正體中文 (繁體)",  # Traditional Chinese
            "zh_Hans": "中文 (简体)",  # Simplified Chinese
            "hr": "Hrvatski",  # Croatian
            "cs": "čeština",  # Czech
            "da": "Dansk",  # Danish
            # "nl": "Nederlands",  # Dutch
            "en": "English",  # English
            "fi": "Suomi",  # Finnish
            "fr": "Français",  # French
            "gl": "Galego",  # Galician
            "de": "Deutsch",  # German
            "el": "Ελληνικά",  # Greek
            "is": "Íslenska",  # Icelandic
            # "id": "Bahasa Indonesia",  # Indonesian
            "ga": "Gaeilge",  # Irish
            "it": "Italiano",  # Italian
            "ja": "日本語",  # Japanese
            "km": "ខ្មែរ",  # Khmer(Central)
            # "ckb": "Soranî",  # Kurdish (Central)
            "lt": "Lietuvių Kalba",  # Lithuanian
            "nb_NO": "Norsk Bokmål",  # Norwegian Bokmål
            "fa": "فارسی",  # Persian
            "pl": "Polski",  # Polish
            "pt_BR": "Português (Brasil)",  # Portuguese Brazil
            "pt_PT": "Português (Portugal)",  # Portuguese Portugal
            # "ro": "Română",  # Romanian
            "ru": "Русский",  # Russian
            "sn": "chiShona",  # Shona
            # "sr_Latn": "Srpska (latinica)",  # Serbian (latin)
            "sk": "Slovenčina",  # Slovak
            "es": "Español",  # Spanish
            "sw": "Kiswahili",  # Swahili
            "sv": "Svenska",  # Swedish
            "ta": "Tamil",    # Tamil
            # "te": "తెలుగు",  # Telugu
            "tr": "Türkçe",  # Turkish
            "uk": "Українська",  # Ukrainian
            "vi": "Tiếng Việt",  # Vietnamese
        }

        # These are the default settings. They will get overwritten when loading from disk
        self.default_settings = {
            "version": self.common.version,
            "connection_type": "bundled",
            "control_port_address": "127.0.0.1",
            "control_port_port": 9051,
            "socks_address": "127.0.0.1",
            "socks_port": 9050,
            "socket_file_path": "/var/run/tor/control",
            "auth_type": "no_auth",
            "auth_password": "",
            "auto_connect": False,
            "use_autoupdate": True,
            "autoupdate_timestamp": None,
            "bridges_enabled": False,
            "bridges_type": "built-in",  # "built-in", "moat", or "custom"
            "bridges_builtin_pt": "obfs4",  # "obfs4", "meek-azure", or "snowflake"
            "bridges_moat": "",
            "bridges_custom": "",
            "bridges_builtin": {},
            "persistent_tabs": [],
            "locale": None,  # this gets defined in fill_in_defaults()
            "theme": 0,
        }
        self._settings = {}
        self.fill_in_defaults()

    def fill_in_defaults(self):
        """
        If there are any missing settings from self._settings, replace them with
        their default values.
        """
        for key in self.default_settings:
            if key not in self._settings:
                self._settings[key] = self.default_settings[key]

        # Choose the default locale based on the OS preference, and fall-back to English
        if self._settings["locale"] is None:
            language_code, encoding = locale.getdefaultlocale()

            # Default to English
            if not language_code:
                language_code = "en_US"

            if language_code == "pt_PT" and language_code == "pt_BR":
                # Portuguese locales include country code
                default_locale = language_code
            else:
                # All other locales cut off the country code
                default_locale = language_code[:2]

            if default_locale not in self.available_locales:
                default_locale = "en"
            self._settings["locale"] = default_locale

    def build_filename(self):
        """
        Returns the path of the settings file.
        """
        return os.path.join(self.common.build_data_dir(), "onionshare.json")

    def load(self):
        """
        Load the settings from file.
        """
        self.common.log("Settings", "load")

        # If the settings file exists, load it
        if os.path.exists(self.filename):
            try:
                self.common.log("Settings", "load", f"Trying to load {self.filename}")
                with open(self.filename, "r") as f:
                    self._settings = json.load(f)
                    self.fill_in_defaults()
            except Exception:
                pass

        # Make sure data_dir exists
        try:
            os.makedirs(self.get("data_dir"), exist_ok=True)
        except Exception:
            pass

    def save(self):
        """
        Save settings to file.
        """
        self.common.log("Settings", "save")
        open(self.filename, "w").write(json.dumps(self._settings, indent=2))
        self.common.log("Settings", "save", f"Settings saved in {self.filename}")

    def get(self, key):
        return self._settings[key]

    def set(self, key, val):
        # If typecasting int values fails, fallback to default values
        if key == "control_port_port" or key == "socks_port":
            try:
                val = int(val)
            except Exception:
                if key == "control_port_port":
                    val = self.default_settings["control_port_port"]
                elif key == "socks_port":
                    val = self.default_settings["socks_port"]

        self._settings[key] = val



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys
import time
import argparse
import threading
from datetime import datetime
from datetime import timedelta

from .common import Common, CannotFindTor
from .web import Web
from .onion import TorErrorProtocolError, TorTooOldEphemeral, TorTooOldStealth, Onion
from .onionshare import OnionShare
from .mode_settings import ModeSettings
from qrcode import QRCode

def main(cwd=None):
    """
    The main() function implements all of the logic that the command-line version of
    onionshare uses.
    """
    common = Common()
    common.display_banner()

    # OnionShare CLI in OSX needs to change current working directory (#132)
    if common.platform == "Darwin":
        if cwd:
            os.chdir(cwd)

    # Parse arguments
    parser = argparse.ArgumentParser(
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=28)
    )
    # Select modes
    parser.add_argument(
        "--receive", action="store_true", dest="receive", help="Receive files"
    )
    parser.add_argument(
        "--website", action="store_true", dest="website", help="Publish website"
    )
    parser.add_argument(
        "--chat", action="store_true", dest="chat", help="Start chat server"
    )
    # Tor connection-related args
    parser.add_argument(
        "--local-only",
        action="store_true",
        dest="local_only",
        default=False,
        help="Don't use Tor (only for development)",
    )
    parser.add_argument(
        "--connect-timeout",
        metavar="SECONDS",
        dest="connect_timeout",
        default=120,
        help="Give up connecting to Tor after a given amount of seconds (default: 120)",
    )
    parser.add_argument(
        "--config",
        metavar="FILENAME",
        default=None,
        help="Filename of custom global settings",
    )
    # Persistent file
    parser.add_argument(
        "--persistent",
        metavar="FILENAME",
        default=None,
        help="Filename of persistent session",
    )
    # General args
    parser.add_argument("--title", metavar="TITLE", default=None, help="Set a title")
    parser.add_argument(
        "--public",
        action="store_true",
        dest="public",
        default=False,
        help="Don't use a private key",
    )
    parser.add_argument(
        "--auto-start-timer",
        metavar="SECONDS",
        dest="autostart_timer",
        default=0,
        help="Start onion service at scheduled time (N seconds from now)",
    )
    parser.add_argument(
        "--auto-stop-timer",
        metavar="SECONDS",
        dest="autostop_timer",
        default=0,
        help="Stop onion service at schedule time (N seconds from now)",
    )
    # Share args
    parser.add_argument(
        "--no-autostop-sharing",
        action="store_true",
        dest="no_autostop_sharing",
        default=False,
        help="Share files: Continue sharing after files have been sent (default is to stop sharing)",
    )
    parser.add_argument(
        "--log-filenames",
        action="store_true",
        dest="log_filenames",
        default=False,
        help="Log file download activity to stdout"
    )
    parser.add_argument(
        "--qr",
        action="store_true",
        dest="qr",
        default=False,
        help="Display a QR code in the terminal for share links",
    )
    # Receive args
    parser.add_argument(
        "--data-dir",
        metavar="data_dir",
        default=None,
        help="Receive files: Save files received to this directory",
    )
    parser.add_argument(
        "--webhook-url",
        metavar="webhook_url",
        default=None,
        help="Receive files: URL to receive webhook notifications",
    )
    parser.add_argument(
        "--disable-text",
        action="store_true",
        dest="disable_text",
        help="Receive files: Disable receiving text messages",
    )
    parser.add_argument(
        "--disable-files",
        action="store_true",
        dest="disable_files",
        help="Receive files: Disable receiving files",
    )
    # Website args
    parser.add_argument(
        "--disable_csp",
        action="store_true",
        dest="disable_csp",
        default=False,
        help="Publish website: Disable the default Content Security Policy header (allows your website to use third-party resources)",
    )
    parser.add_argument(
        "--custom_csp",
        metavar="custom_csp",
        default=None,
        help="Publish website: Set a custom Content Security Policy header",
    )
    # Other
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        help="Log OnionShare errors to stdout, and web errors to disk",
    )
    parser.add_argument(
        "filename",
        metavar="filename",
        nargs="*",
        help="List of files or folders to share",
    )
    args = parser.parse_args()

    filenames = args.filename
    for i in range(len(filenames)):
        filenames[i] = os.path.abspath(filenames[i])

    receive = bool(args.receive)
    website = bool(args.website)
    chat = bool(args.chat)
    local_only = bool(args.local_only)
    connect_timeout = int(args.connect_timeout)
    config_filename = args.config
    persistent_filename = args.persistent
    title = args.title
    public = bool(args.public)
    autostart_timer = int(args.autostart_timer)
    autostop_timer = int(args.autostop_timer)
    autostop_sharing = not bool(args.no_autostop_sharing)
    print_qr = bool(args.qr)
    data_dir = args.data_dir
    webhook_url = args.webhook_url
    disable_text = args.disable_text
    disable_files = args.disable_files
    disable_csp = bool(args.disable_csp)
    custom_csp = args.custom_csp
    log_filenames = bool(args.log_filenames)
    verbose = bool(args.verbose)

    # Verbose mode?
    common.verbose = verbose

    # Re-load settings, if a custom config was passed in
    if config_filename:
        common.load_settings(config_filename)
    else:
        common.load_settings()

    # Mode settings
    if persistent_filename:
        mode_settings = ModeSettings(common, persistent_filename)
        mode_settings.set("persistent", "enabled", True)
    else:
        mode_settings = ModeSettings(common)

    if receive:
        mode = "receive"
    elif website:
        mode = "website"
    elif chat:
        mode = "chat"
    else:
        mode = "share"

    if mode_settings.just_created:
        # This means the mode settings were just created, not loaded from disk
        mode_settings.set("general", "title", title)
        mode_settings.set("general", "public", public)
        mode_settings.set("general", "autostart_timer", autostart_timer)
        mode_settings.set("general", "autostop_timer", autostop_timer)
        mode_settings.set("general", "qr", print_qr)
        if persistent_filename:
            mode_settings.set("persistent", "mode", mode)
        if mode == "share":
            mode_settings.set("share", "autostop_sharing", autostop_sharing)
            mode_settings.set("share", "log_filenames", log_filenames)
        if mode == "receive":
            if data_dir:
                mode_settings.set("receive", "data_dir", data_dir)
            if webhook_url:
                mode_settings.set("receive", "webhook_url", webhook_url)
            mode_settings.set("receive", "disable_text", disable_text)
            mode_settings.set("receive", "disable_files", disable_files)
        if mode == "website":
            if disable_csp and custom_csp:
                print("You cannot disable the CSP and set a custom one. Either set --disable-csp or --custom-csp but not both.")
                sys.exit()
            if disable_csp:
                mode_settings.set("website", "disable_csp", True)
                mode_settings.set("website", "custom_csp", None)
            if custom_csp:
                mode_settings.set("website", "custom_csp", custom_csp)
                mode_settings.set("website", "disable_csp", False)
            mode_settings.set("website", "log_filenames", log_filenames)
    else:
        # See what the persistent mode was
        mode = mode_settings.get("persistent", "mode")

    # In share and website mode, you must supply a list of filenames
    if mode == "share" or mode == "website":
        # Unless you passed in a persistent filename, in which case get the filenames from
        # the mode settings
        if (
            persistent_filename
            and not mode_settings.just_created
            and len(filenames) != 0
        ):
            filenames = mode_settings.get(mode, "filenames")

        else:
            # Make sure filenames given if not using receiver mode
            if len(filenames) == 0:
                if persistent_filename:
                    mode_settings.delete()

                parser.print_help()
                sys.exit()

            # Validate filenames
            valid = True
            for filename in filenames:
                if not os.path.isfile(filename) and not os.path.isdir(filename):
                    print(f"{filename} is not a valid file.")
                    valid = False
                if not os.access(filename, os.R_OK):
                    print(f"{filename} is not a readable file.")
                    valid = False
            if not valid:
                sys.exit()

        # Save the filenames in persistent file
        if persistent_filename:
            mode_settings.set(mode, "filenames", filenames)

    # In receive mode, you must allows either text, files, or both
    if mode == "receive" and disable_text and disable_files:
        print("You cannot disable both text and files")
        sys.exit()

    # Create the Web object
    web = Web(common, False, mode_settings, mode)

    # Start the Onion object
    try:
        onion = Onion(common, use_tmp_dir=True)
    except CannotFindTor:
        print("You must install tor to use OnionShare from the command line")
        if common.platform == "Darwin":
            print("In macOS, you can do this with Homebrew (https://brew.sh):")
            print("  brew install tor")
        sys.exit()

    try:
        onion.connect(
            custom_settings=False,
            config=config_filename,
            connect_timeout=connect_timeout,
            local_only=local_only,
        )
    except KeyboardInterrupt:
        print("")
        sys.exit()
    except Exception:
        sys.exit()

    # Start the onionshare app
    try:
        common.settings.load()

        # Receive mode needs to know the tor proxy details for webhooks
        if mode == "receive":
            if local_only:
                web.proxies = None
            else:
                (socks_address, socks_port) = onion.get_tor_socks_port()
                web.proxies = {
                    "http": f"socks5h://{socks_address}:{socks_port}",
                    "https": f"socks5h://{socks_address}:{socks_port}",
                }

        app = OnionShare(common, onion, local_only, autostop_timer)
        app.choose_port()

        # Delay the startup if a startup timer was set
        if autostart_timer > 0:
            # Can't set a schedule that is later than the auto-stop timer
            if autostop_timer > 0 and autostop_timer < autostart_timer:
                print(
                    "The auto-stop time can't be the same or earlier than the auto-start time. Please update it to start sharing."
                )
                sys.exit()

            app.start_onion_service(mode, mode_settings, False)
            url = f"http://{app.onion_host}"
            schedule = datetime.now() + timedelta(seconds=autostart_timer)
            if mode == "receive":
                print(
                    f"Files sent to you appear in this folder: {mode_settings.get('receive', 'data_dir')}"
                )
                print("")
                print(
                    "Warning: Receive mode lets people upload files to your computer. Some files can potentially take "
                    "control of your computer if you open them. Only open things from people you trust, or if you know "
                    "what you are doing."
                )
                print("")
                if not mode_settings.get("general", "public"):
                    print(
                        f"Give this address and private key to your sender, and tell them it won't be accessible until: {schedule.strftime('%I:%M:%S%p, %b %d, %y')}"
                    )
                    print(f"Private key: {app.auth_string}")
                else:
                    print(
                        f"Give this address to your sender, and tell them it won't be accessible until: {schedule.strftime('%I:%M:%S%p, %b %d, %y')}"
                    )
            else:
                if not mode_settings.get("general", "public"):
                    print(
                        f"Give this address and private key to your recipient, and tell them it won't be accessible until: {schedule.strftime('%I:%M:%S%p, %b %d, %y')}"
                    )
                    print(f"Private key: {app.auth_string}")
                else:
                    print(
                        f"Give this address to your recipient, and tell them it won't be accessible until: {schedule.strftime('%I:%M:%S%p, %b %d, %y')}"
                    )
            print(url)
            if mode_settings.get("general", "qr"):
                qr = QRCode()
                qr.add_data(url)
                print("Onion address as QR code:")
                qr.print_ascii()
                if not mode_settings.get("general", "public"):
                    qr.clear()
                    qr.add_data(app.auth_string)
                    print("Private key as QR code:")
                    qr.print_ascii()
            print("")
            print("Waiting for the scheduled time before starting...")
            app.onion.cleanup(False)
            time.sleep(autostart_timer)
            app.start_onion_service(mode, mode_settings)
        else:
            app.start_onion_service(mode, mode_settings)
    except KeyboardInterrupt:
        print("")
        sys.exit()
    except (TorTooOldEphemeral, TorTooOldStealth, TorErrorProtocolError) as e:
        print("")
        sys.exit()

    if mode == "website":
        # Prepare files to share
        try:
            web.website_mode.set_file_info(filenames)
        except OSError as e:
            print(e.strerror)
            sys.exit(1)

    if mode == "share":
        # Prepare files to share
        print("Compressing files.")
        try:
            web.share_mode.set_file_info(filenames)
        except OSError as e:
            print(e.strerror)
            sys.exit(1)

        # Warn about sending large files over Tor
        if web.share_mode.download_filesize >= 157_286_400:  # 150mb
            print("")
            print("Warning: Sending a large share could take hours")
            print("")

    # Start OnionShare http service in new thread
    t = threading.Thread(target=web.start, args=(app.port,))
    t.daemon = True
    t.start()

    try:  # Trap Ctrl-C
        time.sleep(0.2)

        # start auto-stop timer thread
        if app.autostop_timer > 0:
            app.autostop_timer_thread.start()

        # Build the URL
        url = f"http://{app.onion_host}"

        print("")
        if autostart_timer > 0:
            print("Server started")
        else:
            if mode == "receive":
                print(
                    f"Files sent to you appear in this folder: {mode_settings.get('receive', 'data_dir')}"
                )
                print("")
                print(
                    "Warning: Receive mode lets people upload files to your computer. Some files can potentially take "
                    "control of your computer if you open them. Only open things from people you trust, or if you know "
                    "what you are doing."
                )
                print("")

                if mode_settings.get("general", "public"):
                    print("Give this address to the sender:")
                    print(url)
                else:
                    print("Give this address and private key to the sender:")
                    print(url)
                    print(f"Private key: {app.auth_string}")
            else:
                if mode_settings.get("general", "public"):
                    print("Give this address to the recipient:")
                    print(url)
                else:
                    print("Give this address and private key to the recipient:")
                    print(url)
                    print(f"Private key: {app.auth_string}")
            if mode_settings.get("general", "qr"):
                qr = QRCode()
                qr.add_data(url)
                print("Onion address as QR code:")
                qr.print_ascii()
                if not mode_settings.get("general", "public"):
                    qr.clear()
                    qr.add_data(app.auth_string)
                    print("Private key as QR code:")
                    qr.print_ascii()

        print("")
        print("Press Ctrl+C to stop the server")

        # Wait for app to close
        while t.is_alive():
            if app.autostop_timer > 0:
                # if the auto-stop timer was set and has run out, stop the server
                if not app.autostop_timer_thread.is_alive():
                    if mode == "share":
                        # If there were no attempts to download the share, or all downloads are done, we can stop
                        if not web.share_mode.download_in_progress or web.share_mode.cur_history_id == 0 or web.done:
                            print("Stopped because auto-stop timer ran out")
                            web.stop(app.port)
                            break
                    elif mode == "receive":
                        if (
                            web.receive_mode.cur_history_id == 0
                            or not web.receive_mode.uploads_in_progress
                        ):
                            print("Stopped because auto-stop timer ran out")
                            web.stop(app.port)
                            break
                        web.receive_mode.can_upload = False
                    else:
                        # website or chat mode
                        print("Stopped because auto-stop timer ran out")
                        web.stop(app.port)
                        break
            # Allow KeyboardInterrupt exception to be handled with threads
            # https://stackoverflow.com/questions/3788208/python-threading-ignores-keyboardinterrupt-exception
            time.sleep(0.2)
    except KeyboardInterrupt:
        web.stop(app.port)
    finally:
        # Shutdown
        web.cleanup()
        t.join()
        onion.cleanup()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
OnionShare | https://onionshare.org/

Copyright (C) 2014-2022 Micah Lee, et al. <micah@micahflee.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from PySide6 import QtCore, QtWidgets, QtGui
import qrcode

from . import strings
from .gui_common import GuiCommon


class Alert(QtWidgets.QMessageBox):
    """
    An alert box dialog.
    """

    def __init__(
        self,
        common,
        message,
        icon=QtWidgets.QMessageBox.NoIcon,
        buttons=QtWidgets.QMessageBox.Ok,
        autostart=True,
        title="OnionShare",
    ):
        super(Alert, self).__init__(None)

        self.common = common

        self.common.log("Alert", "__init__")

        self.setWindowTitle(title)
        self.setWindowIcon(QtGui.QIcon(GuiCommon.get_resource_path("images/logo.png")))
        self.setText(message)
        self.setIcon(icon)
        self.setStandardButtons(buttons)

        if autostart:
            self.exec()


class AddFileDialog(QtWidgets.QFileDialog):
    """
    Overridden version of QFileDialog which allows us to select folders as well
    as, or instead of, files. For adding files/folders to share.

    Note that this dialog can't be used in macOS, only in Windows, Linux, and BSD.
    This is because the macOS sandbox requires native dialogs, and this is a Qt5
    dialog.
    """

    def __init__(self, common, *args, **kwargs):
        QtWidgets.QFileDialog.__init__(self, *args, **kwargs)

        self.common = common
        self.common.log("AddFileDialog", "__init__")

        self.setOption(self.Option.DontUseNativeDialog, True)
        self.setOption(self.Option.ReadOnly, True)
        self.setOption(self.Option.ShowDirsOnly, False)
        self.setFileMode(self.FileMode.ExistingFiles)
        tree_view = self.findChild(QtWidgets.QTreeView)
        tree_view.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        list_view = self.findChild(QtWidgets.QListView, "listView")
        list_view.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

    def accept(self):
        self.common.log("AddFileDialog", "accept")
        QtWidgets.QDialog.accept(self)


class MinimumSizeWidget(QtWidgets.QWidget):
    """
    An empty widget with a minimum width and height, just to force layouts to behave
    """

    def __init__(self, width, height):
        super(MinimumSizeWidget, self).__init__()
        self.setMinimumWidth(width)
        self.setMinimumHeight(height)


class Image(qrcode.image.base.BaseImage):
    """
    A custom Image class, for use with the QR Code pixmap.
    """

    def __init__(self, border, width, box_size, *args, **kwargs):
        self.border = border
        self.width = width
        self.box_size = box_size
        size = (width + border * 2) * box_size
        self._image = QtGui.QImage(size, size, QtGui.QImage.Format_RGB16)
        self._image.fill(QtCore.Qt.white)

    def pixmap(self):
        return QtGui.QPixmap.fromImage(self._image)

    def drawrect(self, row, col):
        painter = QtGui.QPainter(self._image)
        painter.fillRect(
            (col + self.border) * self.box_size,
            (row + self.border) * self.box_size,
            self.box_size,
            self.box_size,
            QtCore.Qt.black,
        )

    def save(self, stream, kind=None):
        pass


class QRCodeDialog(QtWidgets.QDialog):
    """
    A dialog showing a QR code.
    """

    def __init__(self, common, title, text):
        super(QRCodeDialog, self).__init__()

        self.common = common

        self.common.log("QrCode", "__init__")

        self.qr_label_title = QtWidgets.QLabel(self)
        self.qr_label_title.setText(title)
        self.qr_label_title.setAlignment(QtCore.Qt.AlignCenter)

        self.qr_label = QtWidgets.QLabel(self)
        self.qr_label.setPixmap(qrcode.make(text, image_factory=Image).pixmap())
        self.qr_label.setScaledContents(True)
        self.qr_label.setFixedSize(350, 350)

        self.setWindowTitle(strings._("gui_qr_code_dialog_title"))
        self.setWindowIcon(QtGui.QIcon(GuiCommon.get_resource_path("images/logo.png")))
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self.qr_label_title)
        layout.addWidget(self.qr_label)

        self.exec()
